"""
plot tools provides fast and elegant solution for plotting quantum experiment data
"""
from typing import Union, Iterable
import pandas as pd
import numpy as np
import plotly.graph_objects as go
import os
from PIL import Image
from datetime import date, datetime

from jupyter_dash import JupyterDash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output, State
from dash import callback_context


def _save_path(data: str, qubit_name: str = 'qubit_default_name',
               default_path: str = 'D:/Scripts/Measurement_automation/data/qubits/',
               spectroscopy_type: str = 'tts'):
    if spectroscopy_type == 'tts':
        spectroscopy = 'two_tone_spectroscopy'
    else:
        spectroscopy = 'single_tone_spectroscopy'

    current_date = str(date.today())

    parent_dir = default_path
    dir_qubit = os.path.join(default_path, str(qubit_name))
    dir_date = os.path.join(dir_qubit, current_date)
    dir_tts = os.path.join(dir_date, spectroscopy)

    if os.path.exists(parent_dir):

        "checking qubit dir existance"
        if not os.path.exists(dir_qubit):
            os.mkdir(dir_qubit)
        else:
            pass

        "checking date dir existance"
        if not os.path.exists(dir_date):
            os.mkdir(dir_date)
        else:
            pass

        "checking tts dir existance"
        if not os.path.exists(dir_tts):
            os.mkdir(dir_tts)
        else:
            pass

        dir_final = os.path.join(dir_tts, data)
        return dir_final
    else:
        return data


class Heatmap:
    """
    Heatmap class provides easy-to-use workspace for building interactive two tone spectroscopy plots
    """

    @staticmethod
    def plot_figure(data: pd.DataFrame = None, *, x: Union[str, np.ndarray, pd.Series, Iterable] = None,
                    y: Union[str, np.ndarray, pd.Series, Iterable] = None,
                    z: Union[str, np.ndarray, pd.Series, Iterable] = None,
                    theme: str = 'dark',
                    cmap: Union[str, Iterable] = ['#ffd200', '#cb2d3e'],
                    title: str = 'Two Tone Spectroscopy',
                    sub_title: str = None,
                    x_axis_title: str = 'Voltages, V',
                    x_axis_size: int = 13,
                    y_axis_title: str = 'Frequencies, GHz',
                    y_axis_size: int = 13,
                    colorbar_text: str = None,
                    logo: bool = False,
                    logo_local=None,
                    y_logo_pos: float = 1.09,
                    x_logo_pos: float = 1,
                    logo_size: float = 0.09,
                    qubit_number: Union[str, float] = None,
                    exponent_mode=False,
                    width=700,
                    height=600
                    ):

        """
        Built a complete plot of the Two Tone Spectroscopy
        :param data: (optional) Pandas data frame with 3 columns (x, y, z)
        :param x: (optional) Pandas series obj | array-like obj. Currents of Two Tone Spectroscopy
        :param y: (optional) Pandas series obj | array-like obj. Frequencies of Two Tone Spectroscopy
        :param z: (optional) Pandas series obj | array-like obj. Response of Two Tone Spectroscopy
        :param theme: ['white', 'dark'] - plot style. default - 'dark'
        :param cmap: The 'colorscale' property is a colorscale and may be specified as:
                      - A list of colors that will be spaced evenly to create the colorscale.
                        Many predefined colorscale lists are included in the sequential, diverging,
                        and cyclical modules in the plotly.colors package.
                      - A list of 2-element lists where the first element is the
                        normalized color level value (starting at 0 and ending at 1),
                        and the second item is a valid color string.
                        (e.g. [[0, 'green'], [0.5, 'red'], [1.0, 'rgb(0, 0, 255)']])
                      - One of the following named colorscales:
                            ['aggrnyl', 'agsunset', 'algae', 'amp', 'armyrose', 'balance',
                             'blackbody', 'bluered', 'blues', 'blugrn', 'bluyl', 'brbg',
                             'brwnyl', 'bugn', 'bupu', 'burg', 'burgyl', 'cividis', 'curl',
                             'darkmint', 'deep', 'delta', 'dense', 'earth', 'edge', 'electric',
                             'emrld', 'fall', 'geyser', 'gnbu', 'gray', 'greens', 'greys',
                             'haline', 'hot', 'hsv', 'ice', 'icefire', 'inferno', 'jet',
                             'magenta', 'magma', 'matter', 'mint', 'mrybm', 'mygbm', 'oranges',
                             'orrd', 'oryel', 'oxy', 'peach', 'phase', 'picnic', 'pinkyl',
                             'piyg', 'plasma', 'plotly3', 'portland', 'prgn', 'pubu', 'pubugn',
                             'puor', 'purd', 'purp', 'purples', 'purpor', 'rainbow', 'rdbu',
                             'rdgy', 'rdpu', 'rdylbu', 'rdylgn', 'redor', 'reds', 'solar',
                             'spectral', 'speed', 'sunset', 'sunsetdark', 'teal', 'tealgrn',
                             'tealrose', 'tempo', 'temps', 'thermal', 'tropic', 'turbid',
                             'turbo', 'twilight', 'viridis', 'ylgn', 'ylgnbu', 'ylorbr',
                             'ylorrd'].
        :param title: (Optional) Sets the plot's title. Default - 'Two Tone Spectroscopy'
        :param sub_title: (Optional) additional information about the plot. Default - None
        :param x_axis_title: (Optional) Sets the title of this axis. Default - 'Currents, A'
        :param x_axis_size: size of x axis font
        :param y_axis_title: (Optional) Sets the title of this axis. Default - 'Frequencies, Hz'
        :param y_axis_size: size of y axis font
        :param colorbar_text: (Optional) Sets the title of color bar. Default - None
        :param logo: (Optional) If True - shows the FMN logo at the upper right corner. Default - True
        :param logo_local: (Optional) If not None - puts <filename> file as a logo
        :param y_logo_pos: sets y logo position
        :param x_logo_pos: sets x logo position
        :param logo_size: logo scale parameter
        :param qubit_number: qubit information
        :param exponent_mode: if False - ignore exponential formatting
        :param width: the width of the plot (in pixels)
        :param height: the height of the plot (in pixels)
        :return:
        """

        if data is not None:
            fig = go.Figure(data=go.Heatmap(
                z=data.iloc[:, 2],
                x=data.iloc[:, 0],
                y=data.iloc[:, 1],
                colorscale=cmap))
        else:
            fig = go.Figure(data=go.Heatmap(
                z=z,
                x=x,
                y=y,
                colorscale=cmap))

        """Adding subtitle"""
        Sub_title = '' if sub_title is None else '<br>' + '<i>' + sub_title + '</i>'
        Title = title + Sub_title if qubit_number is None else title + ' (qubit ' + str(qubit_number) + ')' + Sub_title

        """Choosing colors"""
        if theme.lower() == 'white':
            fig.update_layout(template="plotly_white")
            title_clr = '#000000'
            bg_clr = '#F5F5F5'
            paper_clr = '#FAFAFA'
            grid_clr = '#757575'
        elif theme.lower() == 'dark':
            fig.update_layout(template="plotly_dark")
            title_clr = '#ffffff'
            bg_clr = '#1c1c1c'
            paper_clr = '#1c1c1c'
            grid_clr = '#757575'
        elif theme.lower() == 'pure_white':
            fig.update_layout(template="plotly_white")
            title_clr = '#000000'
            bg_clr = '#ffffff'
            paper_clr = '#ffffff'
            grid_clr = '#757575'
        else:
            fig.update_layout(template="plotly_dark")
            raise NameError

        """layout settings. See more https://plotly.com/python/reference/layout/#layout-title"""

        fig.update_layout(title={
            'text': Title,
            'font': {
                'color': title_clr
            },
            'x': 0.086
        },
            font=dict(family='Open Sans', color=title_clr),
            xaxis_title=x_axis_title,
            yaxis_title=y_axis_title,
            autosize=False,
            separators='.',
            paper_bgcolor=paper_clr,
            plot_bgcolor=bg_clr)

        fig.update_yaxes(title_font={"size": y_axis_size}, tickfont_size=y_axis_size)
        fig.update_xaxes(title_font={"size": x_axis_size}, tickfont_size=x_axis_size)

        """figure size"""
        # if width is not None:
        #     fig.update_layout(width=width)
        # elif height is not None:
        #     fig.update_layout(height=height)
        # elif (width is not None) and (height is not None):
        #     fig.update_layout(width=width, height=height)
        # else:
        #     pass
        fig.update_layout(width=width, height=height)

        """upd numbers formatting. See more https://plotly.com/python/tick-formatting/"""
        if not exponent_mode:
            fig.update_layout(yaxis=dict(showexponent='none', exponentformat='e'))
            fig.update_traces(zhoverformat='.2f')
            # fig.update_layout(xaxis=dict(showexponent='none', exponentformat='e'))
        else:
            fig.update_layout(yaxis_tickformat='.2e', xaxis_tickformat='.2e')
            fig.update_traces(zhoverformat='.2f')

        """add color bar text"""
        if colorbar_text is not None:
            fig.update_traces(colorbar_title_text=colorbar_text)
        else:
            pass

        """adding logo. See more https://plotly.com/python/images/"""
        if logo:
            fig.add_layout_image(
                dict(
                    source='https://raw.githubusercontent.com/ikaryss/pyquac/Master/images/logo_sign.png',
                    xref="paper", yref="paper",
                    x=x_logo_pos, y=y_logo_pos,
                    sizex=logo_size,
                    sizey=logo_size,
                    xanchor="right", yanchor="bottom",
                    opacity=1,
                    layer="above"))
        else:
            pass

        if logo_local is not None:
            img = Image.open('logo_sign.png')
            fig.add_layout_image(
                dict(
                    source=img,
                    xref="paper", yref="paper",
                    x=x_logo_pos, y=y_logo_pos,
                    sizex=logo_size,
                    sizey=logo_size,
                    xanchor="right", yanchor="bottom",
                    opacity=1,
                    layer="above"))
        else:
            pass

        """grid color. See more https://plotly.com/python/axes/"""
        fig.update_layout(xaxis=dict(showgrid=True, gridcolor=grid_clr, gridwidth=0.1,
                                     zeroline=True, zerolinewidth=0.1, zerolinecolor=grid_clr),
                          yaxis=dict(showgrid=True, gridcolor=grid_clr, gridwidth=0.1,
                                     zeroline=True, zerolinewidth=0.1, zerolinecolor=grid_clr))

        return fig

    @staticmethod
    def decorate_figure(fig, *,
                        theme: str = 'dark',
                        cmap: Union[str, Iterable] = ['#ffd200', '#cb2d3e'],
                        title: str = 'Two Tone Spectroscopy',
                        sub_title: str = None,
                        x_axis_title: str = 'Voltages, V',
                        x_axis_size: int = 13,
                        y_axis_title: str = 'Frequencies, GHz',
                        y_axis_size: int = 13,
                        colorbar_text: str = None,
                        logo: bool = False,
                        logo_local=False,
                        y_logo_pos: float = 1.09,
                        x_logo_pos: float = 1,
                        logo_size: float = 0.09,
                        qubit_number: Union[str, float] = None,
                        exponent_mode=False,
                        width=None,
                        height=None
                        ):
        """
        Built a complete plot formatter for your go.Fig object.
        :param fig: plotly.graph_objs._figure.Figure object
        :param theme: ['white', 'dark'] - plot style. default - 'dark'
        :param cmap: The 'colorscale' property is a colorscale and may be specified as:
                      - A list of colors that will be spaced evenly to create the colorscale.
                        Many predefined colorscale lists are included in the sequential, diverging,
                        and cyclical modules in the plotly.colors package.
                      - A list of 2-element lists where the first element is the
                        normalized color level value (starting at 0 and ending at 1),
                        and the second item is a valid color string.
                        (e.g. [[0, 'green'], [0.5, 'red'], [1.0, 'rgb(0, 0, 255)']])
                      - One of the following named colorscales:
                            ['aggrnyl', 'agsunset', 'algae', 'amp', 'armyrose', 'balance',
                             'blackbody', 'bluered', 'blues', 'blugrn', 'bluyl', 'brbg',
                             'brwnyl', 'bugn', 'bupu', 'burg', 'burgyl', 'cividis', 'curl',
                             'darkmint', 'deep', 'delta', 'dense', 'earth', 'edge', 'electric',
                             'emrld', 'fall', 'geyser', 'gnbu', 'gray', 'greens', 'greys',
                             'haline', 'hot', 'hsv', 'ice', 'icefire', 'inferno', 'jet',
                             'magenta', 'magma', 'matter', 'mint', 'mrybm', 'mygbm', 'oranges',
                             'orrd', 'oryel', 'oxy', 'peach', 'phase', 'picnic', 'pinkyl',
                             'piyg', 'plasma', 'plotly3', 'portland', 'prgn', 'pubu', 'pubugn',
                             'puor', 'purd', 'purp', 'purples', 'purpor', 'rainbow', 'rdbu',
                             'rdgy', 'rdpu', 'rdylbu', 'rdylgn', 'redor', 'reds', 'solar',
                             'spectral', 'speed', 'sunset', 'sunsetdark', 'teal', 'tealgrn',
                             'tealrose', 'tempo', 'temps', 'thermal', 'tropic', 'turbid',
                             'turbo', 'twilight', 'viridis', 'ylgn', 'ylgnbu', 'ylorbr',
                             'ylorrd'].
        :param title: (Optional) Sets the plot's title. Default - 'Two Tone Spectroscopy'
        :param sub_title: (Optional) additional information about the plot. Default - None
        :param x_axis_title: (Optional) Sets the title of this axis. Default - 'Currents, A'
        :param y_axis_title: (Optional) Sets the title of this axis. Default - 'Frequencies, Hz'
        :param colorbar_text: (Optional) Sets the title of color bar. Default - None
        :param logo: (Optional) If True - shows the FMN logo at the upper right corner. Default - True
        :param y_logo_pos: y position of the logo. Default = 1.09. Usually belongs to [0.9, 1.1] interval
        """

        """More figure|plot settings here: https://plotly.com/python/reference/layout/"""

        """Adding subtitle"""
        Sub_title = '' if sub_title is None else '<br>' + '<i>' + sub_title + '</i>'
        Title = title + Sub_title if qubit_number is None else title + ' (qubit ' + str(qubit_number) + ')' + Sub_title

        """Choosing colors"""
        if theme.lower() == 'white':
            fig.update_layout(template="plotly_white")
            title_clr = '#000000'
            bg_clr = '#F5F5F5'
            paper_clr = '#FAFAFA'
            grid_clr = '#757575'
        elif theme.lower() == 'dark':
            fig.update_layout(template="plotly_dark")
            title_clr = '#ffffff'
            bg_clr = '#1c1c1c'
            paper_clr = '#1c1c1c'
            grid_clr = '#757575'
        elif theme.lower() == 'pure_white':
            fig.update_layout(template="plotly_white")
            title_clr = '#000000'
            bg_clr = '#ffffff'
            paper_clr = '#ffffff'
            grid_clr = '#757575'
        else:
            fig.update_layout(template="plotly_dark")
            raise NameError

        """Color scale plotly. See more https://plotly.com/python/builtin-colorscales/"""
        fig.update_traces(colorscale=cmap)

        """layout settings. See more https://plotly.com/python/reference/layout/#layout-title"""

        fig.update_layout(title={
            'text': Title,
            'font': {
                'color': title_clr
            },
            'x': 0.086
        },
            font=dict(family='Open Sans', color=title_clr),
            xaxis_title=x_axis_title,
            yaxis_title=y_axis_title,
            autosize=False,
            separators='.',
            paper_bgcolor=paper_clr,
            plot_bgcolor=bg_clr)

        fig.update_yaxes(title_font={"size": y_axis_size}, tickfont_size=y_axis_size)
        fig.update_xaxes(title_font={"size": x_axis_size}, tickfont_size=x_axis_size)

        """figure size"""
        if width is not None:
            fig.update_layout(width=width)
        elif height is not None:
            fig.update_layout(height=height)
        elif (width is not None) and (height is not None):
            fig.update_layout(width=width, height=height)
        else:
            pass

        """upd numbers formatting. See more https://plotly.com/python/tick-formatting/"""
        if not exponent_mode:
            fig.update_layout(yaxis=dict(showexponent='none', exponentformat='e'))
            # fig.update_layout(xaxis=dict(showexponent='none', exponentformat='e'))
            fig.update_traces(zhoverformat='.2f')
        else:
            fig.update_layout(yaxis_tickformat='.2e', xaxis_tickformat='.2e')
            fig.update_traces(zhoverformat='.2f')

        """add color bar text"""
        if colorbar_text is not None:
            fig.update_traces(colorbar_title_text=colorbar_text)
        else:
            pass

        """adding logo. See more https://plotly.com/python/images/"""
        if logo:
            fig.add_layout_image(
                dict(
                    source='https://raw.githubusercontent.com/ikaryss/pyquac/Master/images/logo_sign.png',
                    xref="paper", yref="paper",
                    x=x_logo_pos, y=y_logo_pos,
                    sizex=logo_size,
                    sizey=logo_size,
                    xanchor="right", yanchor="bottom",
                    opacity=1,
                    layer="above"))
        else:
            pass

        if logo_local:
            img = Image.open('logo_sign.png')
            fig.add_layout_image(
                dict(
                    source=img,
                    xref="paper", yref="paper",
                    x=x_logo_pos, y=y_logo_pos,
                    sizex=logo_size,
                    sizey=logo_size,
                    xanchor="right", yanchor="bottom",
                    opacity=1,
                    layer="above"))
        else:
            pass

        """grid color. See more https://plotly.com/python/axes/"""
        fig.update_layout(xaxis=dict(showgrid=True, gridcolor=grid_clr, gridwidth=0.1,
                                     zeroline=True, zerolinewidth=0.1, zerolinecolor=grid_clr),
                          yaxis=dict(showgrid=True, gridcolor=grid_clr, gridwidth=0.1,
                                     zeroline=True, zerolinewidth=0.1, zerolinecolor=grid_clr))

        pass

    """
    info https://plotly.com/python/static-image-export/
    requirements:
    conda install -c conda-forge python-kaleido
    
    """

    @staticmethod
    def save_fig_as(fig, name: str = 'figure.svg', method: str = 'image', create_folder: bool = False,
                    folder_name: str = 'images'):
        """

        :param fig:
        :param name: export format options for images: [.svg (default), .pdf, .png, .jpeg, .webp]
                     HTML export
        :param method: exporting method [image (default), html]
        :param create_folder:
        :param folder_name:
        :return:
        """
        if create_folder == True:
            if not os.path.exists(folder_name):
                os.mkdir(folder_name)

            directory = folder_name + '/' + name

        else:
            directory = name

        if method == 'image':
            fig.write_image(directory)

        elif method == 'html':
            fig.write_html(directory)


class Dash_app:

    def __init__(self, data, fig, qubit_id=None, chip_id=None, additional_info: str = '', spectroscopy_type='tts'):
        """

        :param data: entity of Tts | Sts or TwoToneSpectroscopy class
        :param fig: displayed figure
        :param qubit_id: id of qubit
        :param chip_id: id of chip
        :param spectroscopy_type: tts or sts
        """
        self.data = data
        self.fig = fig
        self.qubit_id = qubit_id
        self.chip_id = chip_id
        self.additional_info = '_' + additional_info + '_'
        self.spectroscopy_type = spectroscopy_type
        self._ts = 'TTS_' if self.spectroscopy_type == 'tts' else 'STS_'

    def __middle_name(self, time):
        return 'q' + str(self.qubit_id) + self.additional_info + str(
            time) if self.qubit_id is not None else 'q' + '_untitled' + self.additional_info + str(time)

    def configure_app(self, mode='inline', port=8051, interval=4e3):
        """
        Configures Dashboard app
        :param mode: 'inline', 'external'
        :param port:
        :param interval: default update time in ms
        :return:
        """
        self.mode = mode
        self.interval = interval
        self.port = port

        disabled_btn = False
        disabled_input = False
        maxx_interval = -1

        # Build App
        app = JupyterDash(__name__)
        app.layout = html.Div(
            [
                html.Button('raw CSV', id='btn-nclicks-1', n_clicks=0),
                html.Button('CSV', id='btn-nclicks-6', n_clicks=0),
                html.Button('PDF', id='btn-nclicks-2', n_clicks=0),
                html.Button('HTML', id='btn-nclicks-3', n_clicks=0),
                html.Button('SVG', id='btn-nclicks-5', n_clicks=0),
                dcc.Checklist(id='checkbox',
                              options=[{'label': 'stop live upd', 'value': 'NO'}],
                              value=['YES', 'NO'],
                              labelStyle={'display': 'inline-block'}
                              ),
                html.Button('Manual upd', id='btn-nclicks-4', n_clicks=0),
                dcc.Input(
                    id="input_time",
                    type='number',
                    placeholder="Auto upd in, ms",
                    min=800,
                ),
                dcc.Graph(id="heatmap", figure=self.fig),
                dcc.Interval(id="animateInterval", interval=self.interval, n_intervals=0,
                             max_intervals=maxx_interval),
            ],
        )

        @app.callback(
            Output("heatmap", "figure"),
            Output("animateInterval", "max_intervals"),
            Output("animateInterval", "interval"),
            Output('btn-nclicks-4', 'disabled'),
            Output('input_time', 'disabled'),
            Input('btn-nclicks-1', 'n_clicks'),
            Input('btn-nclicks-6', 'n_clicks'),
            Input('btn-nclicks-2', 'n_clicks'),
            Input('btn-nclicks-3', 'n_clicks'),
            Input('btn-nclicks-5', 'n_clicks'),
            Input('checkbox', 'value'),
            Input('btn-nclicks-4', 'n_clicks'),
            Input('input_time', 'value'),
            Input("animateInterval", "n_intervals"),
        )
        def doUpdate(btn1, btn6, btn2, btn3, btn5, chkbx, btn4, time_val, i):

            if chkbx[-1] == 'NO':
                maxx_interval = 0
                disabled_btn = False
                disabled_input = False
            else:
                maxx_interval = -1
                disabled_btn = True
                disabled_input = True

            changed_id = [p['prop_id'] for p in callback_context.triggered][0]

            # self.middle_name = 'q' + str(self.qubit_id) + self.additional_info + str(datetime.now().strftime("_%H-%M-%S")) if self.qubit_id is not None else 'q' + '_untitled' + self.additional_info + str(datetime.now().strftime("_%H-%M-%S"))

            # CSV
            if 'btn-nclicks-1' in changed_id:
                file_name = (self._ts + 'raw_' + self.__middle_name(datetime.now().strftime("_%H-%M-%S")) + '.csv')
                self.data.get_result().to_csv(
                    _save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type), index=False)

            # CSV FORMATTED
            elif 'btn-nclicks-6' in changed_id:
                file_name = (self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S")) + '.csv')
                self.data.get_raw_result().to_csv(
                    _save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type), index=False)

            # PDF
            elif 'btn-nclicks-2' in changed_id:
                file_name = (self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S")) + '.pdf')
                fig.write_image(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))

            # HTML
            elif 'btn-nclicks-3' in changed_id:
                file_name = (self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S")) + '.html')
                fig.write_html(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))

            # SVG
            elif 'btn-nclicks-5' in changed_id:
                file_name = (self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S")) + '.svg')
                fig.write_image(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))

            elif 'btn-nclicks-4' in changed_id:
                z = self.data.njit_result
                fig.update_traces(z=z)
            else:
                pass

            if time_val:
                self.interval = time_val

            z = self.data.njit_result

            return self.fig.update_traces(z=z), maxx_interval, self.interval, disabled_btn, disabled_input

        return app.run_server(mode=self.mode, port=self.port)

    def save_all(self,
                 *, csv=True, pdf=True, html=True, svg=True):
        """
        Saves data in different formats
        :param csv: if true - saves csv
        :param pdf: if true - saves pdf
        :param html: if true - saves html
        :param svg: if true - saves svg
        :return:
        """

        # CSV
        if csv:
            file_name = (self._ts + 'raw_' + self.__middle_name(datetime.now().strftime("_%H-%M-%S")) + '.csv')

            self.data.get_result().to_csv(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type),
                                          index=False)

        # CSV FORMATTED
        if csv:
            file_name = (self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S")) + '.csv')
            self.data.get_raw_result().to_csv(
                _save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type), index=False)

        # PDF
        if pdf:
            file_name = (self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S")) + '.pdf')
            fig.write_image(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))

        # HTML
        if html:
            file_name = (self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S")) + '.html')
            fig.write_html(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))

        # SVG
        if svg:
            file_name = (self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S")) + '.svg')
            fig.write_image(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))


class Dash_app_2Q:
    def __init__(self, *,
                 data11, data12, data13,
                 data21, data22, data23,

                 fig11, fig12, fig13,
                 fig21, fig22, fig23,

                 Q1_id=None, Q2_id=None, Coupler_id=None,

                 chip_id=None,

                 check=False,
                 additional_info: str = '', spectroscopy_type='sts'
                 ):
        self.data11, self.data12, self.data13 = data11, data12, data13
        self.data21, self.data22, self.data23 = data21, data22, data23

        self.data_zoo = np.array([self.data11, self.data12, self.data12,
                                  self.data21, self.data22, self.data23])

        self.main_data = self.data11
        self.picked_data = None

        self.fig11, self.fig12, self.fig13 = fig11, fig12, fig13
        self.fig21, self.fig22, self.fig23 = fig21, fig22, fig23

        self.fig_zoo = np.array([self.fig11, self.fig12, self.fig12,
                                 self.fig21, self.fig22, self.fig23])

        self.main_fig = self.fig11
        self.picked_fig = None

        self.Q1_id, self.Q2_id, self.Coupler_id = Q1_id, Q2_id, Coupler_id
        self.chip_id = chip_id

        self.additional_info = '_' + additional_info + '_'
        self.spectroscopy_type = spectroscopy_type
        self._ts = 'STS_' if self.spectroscopy_type == 'sts' else 'TTS_'

        self.check = check

        self.__dict_of_data = dict(fig11=self.data11, fig12=self.data12, fig13=self.data13,
                                   fig21=self.data21, fig22=self.data22, fig23=self.data23)

        self.__dict_of_figs = dict(fig11=self.fig11, fig12=self.fig12, fig13=self.fig13,
                                   fig21=self.fig21, fig22=self.fig22, fig23=self.fig23)
        pass

    def __middle_name(self, time, data):
        if self.check == False:
            mid_name = (data.readout_qubit + '_' + self.__get_id(
                data.readout_qubit.upper()) + '_' + data.flux_qubit + '_ '
                        + self.__get_id(data.flux_qubit.upper()) + '_' + self.additional_info + str(time))
        else:
            mid_name = ('check_' + data.readout_qubit + '_' + self.__get_id(data.readout_qubit.upper()) + '_'
                        + data.flux_qubit + '_ '
                        + self.__get_id(data.flux_qubit.upper()) + '_' + self.additional_info + str(time))
        return mid_name

    def __get_id(self, qubit):
        if qubit == 'Q1':
            qid = self.Q1_id
        elif qubit == 'Q2':
            qid = self.Q2_id
        elif qubit == 'Coupler':
            qid = self.Coupler_id
        else:
            qid = 'Nan'
        return str(qid)

    def configure_app(self, mode='external', port=8051, interval=4e3):
        """
        Configures Dashboard app
        :param mode: 'inline', 'external'
        :param port:
        :param interval: default update time in ms
        :return:
        """
        self.mode = mode
        self.interval = interval
        self.port = port

        disabled_btn = False
        disabled_input = False
        maxx_interval = -1

        # Build App
        app = JupyterDash(__name__)
        app.layout = html.Div(
            [
                html.Button('raw CSV', id='btn-nclicks-1', n_clicks=0),
                html.Button('CSV', id='btn-nclicks-6', n_clicks=0),
                html.Button('PDF', id='btn-nclicks-2', n_clicks=0),
                html.Button('HTML', id='btn-nclicks-3', n_clicks=0),
                html.Button('SVG', id='btn-nclicks-5', n_clicks=0),
                dcc.Dropdown(options=['fig11', 'fig12', 'fig13',
                                      'fig21', 'fig22', 'fig23'], value='fig11', id='fig-dropdown'),

                dcc.Checklist(id='checkbox',
                              options=[{'label': 'stop live upd', 'value': 'NO'}],
                              value=['YES', 'NO'],
                              labelStyle={'display': 'inline-block'}
                              ),
                html.Button('Manual upd', id='btn-nclicks-4', n_clicks=0),
                dcc.Input(
                    id="input_time",
                    type='number',
                    placeholder="Auto upd in, ms",
                    min=800,
                ),

                dcc.Graph(id="heatmap11", figure=self.fig11), dcc.Graph(id="heatmap12", figure=self.fig12),
                dcc.Graph(id="heatmap13", figure=self.fig13), dcc.Graph(id="heatmap21", figure=self.fig21),
                dcc.Graph(id="heatmap22", figure=self.fig22), dcc.Graph(id="heatmap23", figure=self.fig23),

                dcc.Interval(id="animateInterval", interval=self.interval, n_intervals=0,
                             max_intervals=maxx_interval),
            ],
        )

        @app.callback(
            Output("heatmap11", "figure"), Output("heatmap12", "figure"), Output("heatmap13", "figure"),
            Output("heatmap21", "figure"), Output("heatmap22", "figure"), Output("heatmap23", "figure"),
            Output("animateInterval", "max_intervals"),
            Output("animateInterval", "interval"),
            Output('btn-nclicks-4', 'disabled'),
            Output('input_time', 'disabled'),
            Input('btn-nclicks-1', 'n_clicks'),
            Input('btn-nclicks-6', 'n_clicks'),
            Input('btn-nclicks-2', 'n_clicks'),
            Input('btn-nclicks-3', 'n_clicks'),
            Input('btn-nclicks-5', 'n_clicks'),
            Input('fig-dropdown', 'value'),
            Input('checkbox', 'value'),
            Input('btn-nclicks-4', 'n_clicks'),
            Input('input_time', 'value'),
            Input("animateInterval", "n_intervals"),
        )
        def doUpdate(btn1, btn6, btn2, btn3, btn5, fig_dropd_value, chkbx, btn4, time_val, i):

            active = np.array([self.data11.active, self.data12.active, self.data12.active,
                               self.data21.active, self.data22.active, self.data23.active])

            self.picked_data = self.data_zoo[active][0] if len(self.data_zoo[active]) == 1 else self.main_data
            self.picked_fig = self.fig_zoo[active][0] if len(self.fig_zoo[active]) == 1 else self.main_fig

            data_for_save = self.__dict_of_data[fig_dropd_value]
            fig_for_save = self.__dict_of_figs[fig_dropd_value]

            if chkbx[-1] == 'NO':
                maxx_interval = 0
                disabled_btn = False
                disabled_input = False
            else:
                maxx_interval = -1
                disabled_btn = True
                disabled_input = True

            changed_id = [p['prop_id'] for p in callback_context.triggered][0]

            # CSV
            if 'btn-nclicks-1' in changed_id:
                file_name = (self._ts + 'raw_' + self.__middle_name(datetime.now().strftime("_%H-%M-%S"),
                                                                    data_for_save) + '.csv')
                data_for_save.get_result().to_csv(
                    _save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type), index=False)

            # CSV FORMATTED
            elif 'btn-nclicks-6' in changed_id:
                file_name = (
                        self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S"), data_for_save) + '.csv')
                data_for_save.get_raw_result().to_csv(
                    _save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type), index=False)

            # PDF
            elif 'btn-nclicks-2' in changed_id:
                file_name = (
                        self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S"), data_for_save) + '.pdf')
                fig_for_save.write_image(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))

            # HTML
            elif 'btn-nclicks-3' in changed_id:
                file_name = (self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S"),
                                                           data_for_save) + '.html')
                fig_for_save.write_html(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))

            # SVG
            elif 'btn-nclicks-5' in changed_id:
                file_name = (
                        self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S"), data_for_save) + '.svg')
                fig_for_save.write_image(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))

            if time_val:
                self.interval = time_val

            z = self.picked_data.njit_result

            return self.picked_fig.update_traces(z=z), maxx_interval, self.interval, disabled_btn, disabled_input

        return app.run_server(mode=self.mode, port=self.port)


class Dash_app_2Q_base:
    def __init__(self, *,
                 data1,
                 fig,

                 data2=None, data3=None,
                 data4=None, data5=None, data6=None,

                 Q1_id=None, Q2_id=None, Coupler_id=None,

                 chip_id=None,

                 check=False,
                 additional_info: str = '', spectroscopy_type='sts'
                 ):
        self.data11, self.data12, self.data13 = data1, data2, data3
        self.data21, self.data22, self.data23 = data4, data5, data6

        self.data_zoo = np.array([self.data11, self.data12, self.data13,
                                  self.data21, self.data22, self.data23])

        'block for dropdown menu'
        none_count = 0
        for data in self.data_zoo:
            if data is None:
                none_count += 1

        self.existing_data = self.data_zoo[:-none_count]

        self.disabled_drpdwn = False if none_count < 5 else True

        self.data_zoo_names = []
        data_zoo_names_temp = np.array([self.get_label_name(d) for d in self.data_zoo[:-none_count]]) if none_count > 0 \
            else np.array([self.get_label_name(d) for d in self.data_zoo])
        for name in data_zoo_names_temp:
            self.data_zoo_names.append(name)

        while len(self.data_zoo_names) < len(self.data_zoo):
            self.data_zoo_names.append('undefined')
        'end of block for dropdown menu'

        self.data_zoo_names = np.array(self.data_zoo_names)

        self.disabled_markers = []
        for name in self.data_zoo_names:
            if name == 'undefined':
                self.disabled_markers.append(True)
            else:
                self.disabled_markers.append(False)

        self.fig_zoo = {
            self.data_zoo_names[0]: 0, self.data_zoo_names[1]: 1, self.data_zoo_names[2]: 2,
            self.data_zoo_names[3]: 3, self.data_zoo_names[4]: 4, self.data_zoo_names[5]: 5,
        }

        self.main_data = self.data11
        self.picked_data = None

        self.fig = fig

        self.Q1_id, self.Q2_id, self.Coupler_id = Q1_id, Q2_id, Coupler_id
        self.chip_id = chip_id

        self.additional_info = '_' + additional_info + '_'
        self.spectroscopy_type = spectroscopy_type
        self._ts = 'STS_' if self.spectroscopy_type == 'sts' else 'TTS_'

        self.check = check

        self.__dict_of_data = dict(fig11=self.data11, fig12=self.data12, fig13=self.data13,
                                   fig21=self.data21, fig22=self.data22, fig23=self.data23)

        pass

    def __middle_name(self, time, data):
        if not self.check:
            mid_name = (data.readout_qubit + '_' + data.flux_qubit + '_'
                        + 'sweep' + self.additional_info + str(time))
        else:
            mid_name = ('check_' + data.readout_qubit + '_' + data.flux_qubit + '_'
                        + 'sweep' + self.additional_info + str(time))
        return mid_name

    def __get_id(self, qubit):
        if qubit == 'Q1':
            qid = self.Q1_id
        elif qubit == 'Q2':
            qid = self.Q2_id
        elif qubit == 'Coupler':
            qid = self.Coupler_id
        else:
            qid = 'Nan'
        return str(qid)

    def get_label_name(self, data):
        readout = data.readout_qubit
        flux = data.flux_qubit
        if isinstance(flux, str):
            return str(readout) + '_' + str(flux) + '_sweep'
        else:
            global_str = str(readout) + '_'
            for f in flux:
                global_str += str(f) + '+'
            global_str = global_str[:-1]
            global_str += '_sweep'
            return global_str

    def configure_app(self, mode='inline', port=8051, interval=4e3):
        """
        Configures Dashboard app
        :param mode: 'inline', 'external'
        :param port:
        :param interval: default update time in ms
        :return:
        """
        self.mode = mode
        self.interval = interval
        self.port = port

        disabled_btn = False
        disabled_input = False
        maxx_interval = -1

        # Build App
        app = JupyterDash(__name__)
        app.layout = html.Div(
            [
                html.Button('raw CSV', id='btn-nclicks-1', n_clicks=0),
                html.Button('CSV', id='btn-nclicks-6', n_clicks=0),
                html.Button('PDF', id='btn-nclicks-2', n_clicks=0),
                html.Button('HTML', id='btn-nclicks-3', n_clicks=0),
                html.Button('SVG', id='btn-nclicks-5', n_clicks=0),
                dcc.Dropdown(
                    options=[{'label': value, 'value': value, 'disabled': disabl} for value, disabl in
                             zip(self.data_zoo_names, self.disabled_markers)],
                    value=self.get_label_name(self.data11), id='fig-dropdown',
                    disabled=self.disabled_drpdwn,
                    clearable=False
                ),

                dcc.Checklist(id='checkbox',
                              options=[{'label': 'stop live upd', 'value': 'NO'}],
                              value=['YES', 'NO'],
                              labelStyle={'display': 'inline-block'}
                              ),
                html.Button('Manual upd', id='btn-nclicks-4', n_clicks=0),
                dcc.Input(
                    id="input_time",
                    type='number',
                    placeholder="Auto upd in, ms",
                    min=800,
                ),

                dcc.Graph(id="heatmap", figure=self.fig),

                dcc.Interval(id="animateInterval", interval=self.interval, n_intervals=0,
                             max_intervals=maxx_interval),
            ],
        )

        @app.callback(
            Output("heatmap", "figure"),
            Output("animateInterval", "max_intervals"),
            Output("animateInterval", "interval"),
            Output('btn-nclicks-4', 'disabled'),
            Output('input_time', 'disabled'),
            Input('btn-nclicks-1', 'n_clicks'),
            Input('btn-nclicks-6', 'n_clicks'),
            Input('btn-nclicks-2', 'n_clicks'),
            Input('btn-nclicks-3', 'n_clicks'),
            Input('btn-nclicks-5', 'n_clicks'),
            Input('fig-dropdown', 'value'),
            Input('checkbox', 'value'),
            Input('btn-nclicks-4', 'n_clicks'),
            Input('input_time', 'value'),
            Input("animateInterval", "n_intervals"),
        )
        def doUpdate(btn1, btn6, btn2, btn3, btn5, fig_dropd_value, chkbx, btn4, time_val, i):

            self.picked_data = self.data_zoo[
                self.fig_zoo[fig_dropd_value]] if self.disabled_drpdwn is False else self.data11

            if chkbx[-1] == 'NO':
                maxx_interval = 0
                disabled_btn = False
                disabled_input = False
            else:
                maxx_interval = -1
                disabled_btn = True
                disabled_input = True

            changed_id = [p['prop_id'] for p in callback_context.triggered][0]

            # CSV
            if 'btn-nclicks-1' in changed_id:
                file_name = (self._ts + 'raw_' + self.__middle_name(datetime.now().strftime("_%H-%M-%S"),
                                                                    self.picked_data) + '.csv')
                self.picked_data.get_result().to_csv(
                    _save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type), index=False)

            # CSV FORMATTED
            elif 'btn-nclicks-6' in changed_id:
                file_name = (
                        self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S"), self.picked_data) + '.csv')
                self.picked_data.get_raw_result().to_csv(
                    _save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type), index=False)

            # PDF
            elif 'btn-nclicks-2' in changed_id:
                file_name = (
                        self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S"), self.picked_data) + '.pdf')
                self.fig.write_image(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))

            # HTML
            elif 'btn-nclicks-3' in changed_id:
                file_name = (self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S"),
                                                           self.picked_data) + '.html')
                self.fig.write_html(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))

            # SVG
            elif 'btn-nclicks-5' in changed_id:
                file_name = (
                        self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S"), self.picked_data) + '.svg')
                self.fig.write_image(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))

            if time_val:
                self.interval = time_val

            z = self.picked_data.njit_result
            x = self.picked_data.x_1d
            y = self.picked_data.y_1d

            return self.fig.update_traces(z=z, x=x, y=y), maxx_interval, self.interval, disabled_btn, disabled_input

        return app.run_server(mode=self.mode, port=self.port)

    def save_all(self):

        for data in self.existing_data:
            # CSV
            file_name = (self._ts + 'raw_' + self.__middle_name(datetime.now().strftime("_%H-%M-%S"),
                                                                data) + '.csv')
            self.picked_data.get_result().to_csv(
                _save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type), index=False)

            # CSV FORMATTED
            file_name = (
                    self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S"), data) + '.csv')
            self.picked_data.get_raw_result().to_csv(
                _save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type), index=False)

            # PDF
            file_name = (
                    self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S"), data) + '.pdf')
            self.fig.write_image(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))

            # HTML
            file_name = (self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S"),
                                                       data) + '.html')
            self.fig.write_html(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))

            # SVG
            file_name = (
                    self._ts + self.__middle_name(datetime.now().strftime("_%H-%M-%S"), data) + '.svg')
            self.fig.write_image(_save_path(file_name, self.chip_id, spectroscopy_type=self.spectroscopy_type))


class Dash_app_2Q1111:
    def __init__(self, *,
                 data11, data12, data13,
                 data21, data22, data23,

                 fig11, fig12, fig13,
                 fig21, fig22, fig23,

                 Q1_id=None, Q2_id=None, Coupler_id=None,

                 chip_id=None,

                 check=False,
                 additional_info: str = '', spectroscopy_type='sts'
                 ):
        self.data11, self.data12, self.data13 = data11, data12, data13
        self.data21, self.data22, self.data23 = data21, data22, data23

        self.fig11, self.fig12, self.fig13 = fig11, fig12, fig13
        self.fig21, self.fig22, self.fig23 = fig21, fig22, fig23

        self.Q1_id, self.Q2_id, self.Coupler_id = Q1_id, Q2_id, Coupler_id
        self.chip_id = chip_id

        self.additional_info = '_' + additional_info + '_'
        self.spectroscopy_type = spectroscopy_type
        self._ts = 'STS_' if self.spectroscopy_type == 'sts' else 'TTS_'

        self.check = check

        pass

    def configure_app(self, mode='external', port=8051, interval=4e3):
        """
        Configures Dashboard app
        :param mode: 'inline', 'external'
        :param port:
        :param interval: default update time in ms
        :return:
        """
        self.mode = mode
        self.interval = interval
        self.port = port

        external_stylesheets = [
            # Dash CSS
            'https://codepen.io/chriddyp/pen/bWLwgP.css',
            # Loading screen CSS
            'https://codepen.io/chriddyp/pen/brPBPO.css']

        app = JupyterDash(__name__, external_stylesheets=external_stylesheets)
        server = app.server

        CACHE_CONFIG = {
            # try 'FileSystemCache' if you don't want to setup redis
            'CACHE_TYPE': 'redis',
            'CACHE_REDIS_URL': os.environ.get('REDIS_URL', 'redis://localhost:6379')
        }
        cache = Cache()
        cache.init_app(app.server, config=CACHE_CONFIG)

        app.layout = html.Div([
            html.Div([
                html.Div(dcc.Graph(id='heatmap11', figure=self.fig11), className="six columns"),
                html.Div(dcc.Graph(id='heatmap12', figure=self.fig12), className="six columns"),
                html.Div(dcc.Graph(id='heatmap13', figure=self.fig13), className="six columns"),
            ], className="row"),
            html.Div([
                html.Div(dcc.Graph(id='heatmap21', figure=self.fig21), className="six columns"),
                html.Div(dcc.Graph(id='heatmap22', figure=self.fig22), className="six columns"),
                html.Div(dcc.Graph(id='heatmap23', figure=self.fig23), className="six columns"),
            ], className="row"),

            # signal value to trigger callbacks
            dcc.Store(id='signal'),
            dcc.Interval(id="animateInterval", interval=self.interval, n_intervals=0)
        ])

        # perform expensive computations in this "global store"
        # these computations are cached in a globally available
        # redis memory store which is available across processes
        # and for all time.
        @cache.memoize()
        def global_store():
            z11 = self.data11.njit_result
            z12 = self.data12.njit_result
            z13 = self.data13.njit_result

            z21 = self.data21.njit_result
            z22 = self.data22.njit_result
            z23 = self.data23.njit_result

            return z11, z12, z13, z21, z22, z23

        @app.callback(Output('signal', 'data'), Input("animateInterval", "n_intervals"))
        def compute_value(i):
            # compute value and send a signal when done
            global_store()
            return value



from . import datatools
from . import fmn_plottools
from . import fmn_datatools
from . import fmn_tts
from . import fmn_spectroscopy

__all__ = [
    'datatools',
    'fmn_plottools',
    'fmn_datatools',
    'fmn_tts',
    'fmn_spectroscopy'
]



# built-in libraries
from time import perf_counter
from typing import Iterable, Union
import time
# installable libraries
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from scipy import stats
import numba as nb

# working with 2nd qubit
from scipy.optimize import curve_fit

# devices
import zhinst
import zhinst.ziPython
from drivers.M9290A import *
from drivers.N5183B import *
from drivers.znb_besedin import *
from resonator_tools import circlefit
from resonator_tools.circuit import notch_port

# family class
from pyquac.fmn_datatools import TwoToneSpectroscopy, timer


def _fit_cos(t, A1, A0, omega, teta):
    return A1 * np.cos(2 * np.pi * omega * t + teta) + A0


class Tts(TwoToneSpectroscopy):
    """
    Two Tone Spectroscopy for 1 qubit measurements
    """

    def __init__(self,
                 *, x_min, x_max, y_min, y_max,
                 fr_min: float, fr_max: float,
                 nx_points=None, x_step=None,
                 y_step=None, ny_points=None,
                 hdawg_port: str = '127.0.0.1', hdawg_port1: int = 8004, hdawg_mode: int = 6,
                 hdawg_device: str = 'dev8210', hdawg_channel: int = 5,
                 LO_port: str = 'TCPIP0::192.168.180.143::hislip0::INSTR',
                 LO_res_port: str = 'TCPIP0::192.168.180.110::inst0::INSTR',
                 LO_set_power: int = 5,
                 LO_res_set_bandwidth: int = 20, LO_res_set_power: int = -10, LO_res_set_nop=101,
                 base_bandwidth=40, LO_res_set_averages=1, LO_res_meas_averages=1
                 ):
        """
        Class provides methods for working with live data for Two Tone Spectroscopy
        :param x_min: x minimum value (int | float)
        :param x_max: x maximum value (int | float)
        :param y_min: y minimum value (int | float)
        :param y_max: y maximum value (int | float)
        :param nx_points: x count value (int)
        :param x_step: x step value (float)
        :param ny_points: y count value (int)
        :param y_step: y step value (float)
        :param fr_min: min frequency for find resonator
        :param fr_max: max frequency for find resonator
        :param hdawg_port: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_mode)
        :param hdawg_port1: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_mode)
        :param hdawg_mode: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_mode)
        :param LO_port: qubit LO = N5183B('N5183B', LO_port)
        :param LO_res_port: resonator LO_res = Znb(LO_res_port)
        :param hdawg_device: 'dev8210' by default
        :param hdawg_channel: hdawg.setInt('/' + hdawg_device + '/sigouts/' + str(hdawg_channel) + '/on', 1)
        :param LO_set_power: base LO power (default 5)
        :param LO_res_set_bandwidth: bandwidth during resonator tuning
        :param LO_res_set_power: base resonator LO power (default -10)
        :param LO_res_set_nop: number of points during resonator scanning (default 101)
        :param base_bandwidth: (int) bandwidth during mesurments
        :param LO_res_set_averages: set averages for resonator LO parameter
        :param LO_res_meas_averages: measure averages for resonator LO parameter
        """

        super().__init__(x_min=x_min, x_max=x_max, nx_points=nx_points, y_min=y_min, y_max=y_max, ny_points=ny_points,
                         x_step=x_step, y_step=y_step)

        self.fr_min = fr_min
        self.fr_max = fr_max
        self.hdawg_channel = hdawg_channel
        self.hdawg_device = hdawg_device

        # HDAWG init
        self.hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_mode)
        self.hdawg_setDouble = '/' + self.hdawg_device + '/sigouts/' + str(self.hdawg_channel) + '/offset'
        # open HDAWG ZI
        hdawgModule = self.hdawg.awgModule()

        # freq generator init
        self.LO = N5183B('N5183B', LO_port)  # qubit
        self.LO_res = Znb(LO_res_port)  # resonator

        # set base parameters
        self.LO_set_power = LO_set_power
        self.LO_res_set_nop = LO_res_set_nop
        self.LO_res_set_bandwidth = LO_res_set_bandwidth
        self.LO_res_set_power = LO_res_set_power
        self.base_bandwidth = base_bandwidth
        self.LO_res_set_averages = LO_res_set_averages
        self.LO_res.set_averages(LO_res_set_averages)
        self.LO_res_meas_averages = LO_res_meas_averages

        pass

    @property
    def __find_resonator(self):

        self.LO.set_status(0)

        # prior bandwidth
        bandwidth = self.LO_res.get_bandwidth()
        xlim = self.LO_res.get_freq_limits()

        self.LO_res.set_bandwidth(self.LO_res_set_bandwidth)
        self.LO_res.set_nop(self.LO_res_set_nop)
        self.LO_res.set_freq_limits(self.fr_min, self.fr_max)
        self.LO_res.set_averages(self.LO_res_set_averages)

        # measure S21
        freqs = self.LO_res.get_freqpoints()
        notch = notch_port(freqs, self.LO_res.measure()['S-parameter'])
        notch.autofit(electric_delay=60e-9)
        result = round(notch.fitresults['fr'])

        # Resetting to the next round of measurements
        self.LO_res.set_bandwidth(bandwidth)
        self.LO_res.set_freq_limits(*xlim)
        self.LO_res.set_nop(1)
        self.LO.set_status(1)

        return result

    def run_measurements(self, *, x_key: Union[float, int, Iterable] = None, y_key: Union[float, int, Iterable] = None,
                         x_min=None, x_max=None, y_min=None, y_max=None,
                         sleep=0.007, timeout=None):

        # Turn on HDAWG
        self.hdawg.setInt('/' + self.hdawg_device + '/sigouts/' + str(self.hdawg_channel) + '/on', 1)

        self.iter_setup(x_key=x_key, y_key=y_key,
                        x_min=x_min, x_max=x_max, y_min=y_min, y_max=y_max)

        # Set power
        self.LO.set_power(self.LO_set_power)
        self.LO_res.set_power(self.LO_res_set_power)
        # base bandwidth
        self.LO_res.set_bandwidth(self.base_bandwidth)
        try:
            for i in range(len(self.load)):
                if (i == 0) or (self.load[i] != self.load[i - 1]):
                    self.LO_res.set_center(float(self.__find_resonator))

                    if timeout is not None:
                        timer.sleep(timeout)
                    else:
                        pass
                # measurement averages
                self.LO_res.set_averages(self.LO_res_meas_averages)

                self.hdawg.setDouble(self.hdawg_setDouble, self.load[i])  # Current write
                self.LO.set_frequency(self.frequency[i])  # Frequency write

                result = self.LO_res.measure()['S-parameter']

                self.write(x=self.load[i],
                           y=self.frequency[i],
                           heat=20 * np.log10(abs(result)[0])
                           )
                timer.sleep(sleep)

        except KeyboardInterrupt:
            pass

        # Turn off LO
        self.LO.set_status(0)

        # Turn off HDAWG
        self.hdawg.setInt('/' + self.hdawg_device + '/sigouts/' + str(self.hdawg_channel) + '/on', 0)


class Sts(TwoToneSpectroscopy):
    """
    Single Tone Spectroscopy for 1 qubit measurements
    """

    def __init__(self,
                 *, x_min, x_max, y_min, y_max,
                 nx_points=None, x_step=None, y_step=None, ny_points=None,
                 hdawg_port: str = '127.0.0.1', hdawg_port1: int = 8004, hdawg_mode: int = 6,
                 hdawg_channel: int = 5, hdawg_device: str = 'dev8210',
                 LO_res_port: str = 'TCPIP0::192.168.180.110::inst0::INSTR',
                 LO_res_set_bandwidth: int = 20, LO_res_set_power: int = -10,
                 LO_res_set_averages=1
                 ):
        """
        :param x_min: x minimum value (int | float)
        :param x_max: x maximum value (int | float)
        :param y_min: y minimum value (int | float)
        :param y_max: y maximum value (int | float)
        :param x_step: x step value (float)
        :param nx_points: x count value (int)
        :param y_step: y step value (float)
        :param ny_points: y count value (int)
        :param hdawg_port: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_mode)
        :param hdawg_port1: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_mode)
        :param hdawg_mode: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_mode)
        :param LO_res_port: resonator LO_res = Znb(LO_res_port)
        :param hdawg_channel: hdawg.setInt('/' + hdawg_device + '/sigouts/' + str(hdawg_channel) + '/on', 1)
        :param hdawg_device: 'dev8210' by default
        :param LO_res_set_bandwidth: base bandwidth (default 20)
        :param LO_res_set_power: base LO_res power (default -10)
        :param LO_res_set_averages: set averages for resonator LO parameter
        """

        super().__init__(x_min=x_min, x_max=x_max, nx_points=nx_points, y_min=y_min, y_max=y_max, ny_points=ny_points,
                         x_step=x_step, y_step=y_step)

        self.ny_points = ny_points if self.y_step is None else len(self.y_list)

        # HDAWG init
        self.hdawg_device = hdawg_device
        self.hdawg_channel = hdawg_channel

        # HDAWG init
        self.hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_mode)
        hdawgModule = self.hdawg.awgModule()

        # LO connect
        self.LO_res = Znb(LO_res_port)  # resonator

        # set base parameters of LO res
        self.LO_res_set_bandwidth = LO_res_set_bandwidth
        self.LO_res_set_power = LO_res_set_power

        self.LO_res_set_averages = LO_res_set_averages
        self.LO_res.set_averages(LO_res_set_averages)

        pass

    def run_measurements(self, *, sleep=0.0007):

        self.iter_setup(x_key=None, y_key=None,
                        x_min=None, x_max=None, y_min=None, y_max=None)

        # enable channel
        self.hdawg.setInt('/' + self.hdawg_device + '/sigouts/' + str(self.hdawg_channel) + '/on', 1)

        # Set power
        self.LO_res.set_power(self.LO_res_set_power)

        # Set LO parameters
        self.LO_res.set_bandwidth(self.LO_res_set_bandwidth)
        self.LO_res.set_nop(self.ny_points)
        self.LO_res.set_freq_limits(self.y_min, self.y_max)

        freq_len = len(self.y_list)
        try:
            for i in range(len(self.load)):
                if (i == 0) or (self.load[i] != self.load[i - 1]):
                    self.hdawg.setDouble('/' + hdawg_device + '/sigouts/' + str(self.hdawg_channel)
                                         + '/offset', self.load[i])  # current write

                    self.LO_res.set_averages(self.LO_res_set_averages)
                    result = self.LO_res.measure()['S-parameter']

                    for j in range(freq_len):
                        self.write(x=self.load[i],
                                   y=self.y_list[j],
                                   heat=20 * np.log10(abs(result[j]))
                                   )

                timer.sleep(sleep)

        except KeyboardInterrupt:
            if (self.x_raw[-1] == self.x_list[-1]) and (self.y_raw[-1] == self.y_list[-1]):
                pass
            else:
                # drop the last column if interrupt for stable data saving
                self.drop(x=self.x_raw[-1])
            pass

        # channel switch off
        self.hdawg.setInt('/' + self.hdawg_device + '/sigouts/' + str(self.hdawg_channel) + '/on', 0)


class Sts2Q(TwoToneSpectroscopy):
    """
    Single Tone Spectroscopy for 2 qubit measurements
    """

    def __init__(self,
                 *,
                 flux_qubit: str, readout_qubit: str,
                 x_min=None, x_max=None, y_min=None, y_max=None,
                 nx_points=None, x_step=None, y_step=None, ny_points=None,
                 x_min_coupler=None, x_max_coupler=None, x_step_coupler=None, nx_points_coupler=None,
                 x_min_control=None, x_max_control=None, x_step_control=None, nx_points_control=None,
                 target_ch: int = None,
                 control_ch: int = None,
                 coupler_ch: int = None,
                 hdawg_port: str = '127.0.0.1', hdawg_port1: int = 8004, hdawg_mode: int = 6,
                 hdawg_device: str = 'dev8210',
                 LO_res_port: str = 'TCPIP0::192.168.180.110::inst0::INSTR',
                 LO_res_set_bandwidth: int = 20, LO_res_set_power: int = -10,
                 LO_res_set_averages=1
                 ):
        """

        :param flux_qubit: str|list of strings. names of qubits that we drive
        :param readout_qubit: str. name of readout qubit
        :param x_min: min x value of target qubit or base x_min while checking
        :param x_max: max x value of target qubit or base x_max while checking
        :param y_min: min y value of readout qubit
        :param y_max: max y value of readout qubit
        :param nx_points: nx_points value of target qubit or base nx_points while checking
        :param x_step: x_step value of target qubit or base x_step while checking
        :param y_step: y_step value of readout qubit
        :param ny_points: ny_points value of readout qubit
        :param x_min_coupler: min x value of coupler qubit
        :param x_max_coupler: max x value of coupler qubit
        :param x_step_coupler: x_step value of coupler qubit
        :param nx_points_coupler: nx_points value of coupler qubit
        :param x_min_control: min x value of control qubit
        :param x_max_control: max x value of control qubit
        :param x_step_control: x_step value of control qubit
        :param nx_points_control: nx_points value of control qubit
        :param target_ch: target channel
        :param control_ch: control channel
        :param coupler_ch: coupler channel
        :param hdawg_port: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_mode)
        :param hdawg_port1: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_mode)
        :param hdawg_mode: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_mode)
        :param LO_res_port: resonator LO_res = Znb(LO_res_port)
        :param hdawg_channel: hdawg.setInt('/' + hdawg_device + '/sigouts/' + str(hdawg_channel) + '/on', 1)
        :param hdawg_device: 'dev8210' by default
        :param LO_res_set_bandwidth: base bandwidth (default 20)
        :param LO_res_set_power: base LO_res power (default -10)
        :param LO_res_set_averages: set averages for resonator LO parameter
        """

        if (target_ch is not None) and (control_ch is None) and (coupler_ch is None):
            super().__init__(x_min=x_min, x_max=x_max, nx_points=nx_points, y_min=y_min, y_max=y_max,
                             ny_points=ny_points, x_step=x_step, y_step=y_step)
            self.ny_points = ny_points if self.y_step is None else len(self.y_list)
            # channel define
            self.target_ch = target_ch

        elif (coupler_ch is not None) and (control_ch is None) and (target_ch is None):
            super().__init__(x_min=x_min_coupler, x_max=x_max_coupler, nx_points=nx_points_coupler, y_min=y_min,
                             y_max=y_max,
                             ny_points=ny_points, x_step=x_step_coupler, y_step=y_step)
            self.ny_points = ny_points if self.y_step is None else len(self.y_list)
            # channel define
            self.coupler_ch = coupler_ch

        elif (control_ch is not None) and (coupler_ch is None) and (target_ch is None):
            super().__init__(x_min=x_min_control, x_max=x_max_control, nx_points=nx_points_control, y_min=y_min,
                             y_max=y_max, ny_points=ny_points, x_step=x_step_control,
                             y_step=y_step)
            self.ny_points = ny_points if self.y_step is None else len(self.y_list)
            # channel define
            self.control_ch = control_ch

        else:

            if target_ch is not None:
                self.target_ch = target_ch
            if coupler_ch is not None:
                self.coupler_ch = coupler_ch
            if control_ch is not None:
                self.control_ch = control_ch

            super().__init__(x_min=x_min, x_max=x_max, nx_points=nx_points, y_min=y_min,
                             y_max=y_max, ny_points=ny_points, x_step=x_step, y_step=y_step)

            self.ny_points = ny_points if self.y_step is None else len(self.y_list)

        # HDAWG init
        self.hdawg_device = hdawg_device
        self.hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_mode)
        hdawgModule = self.hdawg.awgModule()

        # LO connect
        self.LO_res = Znb(LO_res_port)  # resonator

        # set base parameters of LO res
        self.LO_res_set_bandwidth = LO_res_set_bandwidth
        self.LO_res_set_power = LO_res_set_power

        self.LO_res_set_averages = LO_res_set_averages
        self.LO_res.set_averages(LO_res_set_averages)

        self.flux_qubit = flux_qubit
        self.readout_qubit = readout_qubit

        self.finished = False
        self.active = False
        self.start_for_fit_LP = (6, 0.02, 0.1, 0)
        self.start_for_fit_HP = (10, 0.2, 0.5, 0)

        self.fit = None
        self.x_offset = None
        pass

    def run_measurements(self, *, sleep=0.0007):

        self.iter_setup(x_key=None, y_key=None,
                        x_min=None, x_max=None, y_min=None, y_max=None)

        self.active = True

        # channel switch on
        if self.target_ch is not None:
            self.hdawg.setInt('/' + self.hdawg_device + '/sigouts/' + str(self.target_ch) + '/on', 1)

        if self.coupler_ch is not None:
            self.hdawg.setInt('/' + self.hdawg_device + '/sigouts/' + str(self.coupler_ch) + '/on', 1)

        if self.control_ch is not None:
            self.hdawg.setInt('/' + self.hdawg_device + '/sigouts/' + str(self.control_ch) + '/on', 1)

        # Set power
        self.LO_res.set_power(self.LO_res_set_power)

        # Set LO parameters
        self.LO_res.set_bandwidth(self.LO_res_set_bandwidth)
        self.LO_res.set_nop(self.ny_points)
        self.LO_res.set_freq_limits(self.y_min, self.y_max)

        freq_len = len(self.y_list)
        try:
            for i in range(len(self.load)):
                if (i == 0) or (self.load[i] != self.load[i - 1]):

                    if self.target_ch is not None:
                        self.hdawg.setDouble('/' + self.hdawg_device + '/sigouts/'
                                             + str(self.target_ch) + '/offset', self.load[i])

                    if self.coupler_ch is not None:
                        self.hdawg.setDouble('/' + self.hdawg_device + '/sigouts/'
                                             + str(self.coupler_ch) + '/offset', self.load[i])

                    if self.control_ch is not None:
                        self.hdawg.setDouble('/' + self.hdawg_device + '/sigouts/'
                                             + str(self.control_ch) + '/offset', self.load[i])

                    self.LO_res.set_averages(self.LO_res_set_averages)
                    result = self.LO_res.measure()['S-parameter']

                    for j in range(freq_len):
                        self.write(x=self.load[i],
                                   y=self.y_list[j],
                                   heat=20 * np.log10(abs(result[j]))
                                   )
                timer.sleep(sleep)

        except KeyboardInterrupt:
            if (self.x_raw[-1] == self.x_list[-1]) and (self.y_raw[-1] == self.y_list[-1]):
                pass
            else:
                # drop the last column if interrupt for stable data saving
                self.drop(x=self.x_raw[-1])
            self.active = False
            pass

        if (self.x_raw[-1] == self.x_list[-1]) and (self.y_raw[-1] == self.y_list[-1]):
            self.finished = True
        else:
            pass

        # channel switch off
        if self.target_ch is not None:
            self.hdawg.setInt('/' + self.hdawg_device + '/sigouts/' + str(self.target_ch) + '/on', 0)

        if self.coupler_ch is not None:
            self.hdawg.setInt('/' + self.hdawg_device + '/sigouts/' + str(self.coupler_ch) + '/on', 0)

        if self.control_ch is not None:
            self.hdawg.setInt('/' + self.hdawg_device + '/sigouts/' + str(self.control_ch) + '/on', 0)

        self.active = False

    def fit_and_offset(self, start, fitted_freqs=None):

        if fitted_freqs is not None:
            fitted_params, cov = curve_fit(_fit_cos, self.x_list, fitted_freqs, start)
        else:
            fitted_freqs = self.fit_curve()['y']
            fitted_params, cov = curve_fit(_fit_cos, self.x_list, fitted_freqs, start)

        A1, A0, omega, teta = fitted_params

        fit = A1 * np.cos(2 * np.pi * omega * self.x_list + teta) + A0
        fit_dict = dict(x=self.x_list, y=fit)
        x_offset = self.x_list[np.argmax(fit)]

        self.fit = fit
        self.x_offset = x_offset

        return fit_dict, x_offset


class Sts2QContainer:

    def __init__(self, *,
                 y_minQ1, y_maxQ1, y_minQ2, y_maxQ2,
                 q1_ch: int,
                 q2_ch: int,
                 coupler_ch: int,
                 x_min=None, x_max=None, nx_points=None, x_step=None,

                 x_minQ1=None, x_maxQ1=None, nx_pointsQ1=None, x_stepQ1=None,
                 ny_pointsQ1=None, y_stepQ1=None,

                 x_minC=None, x_maxC=None, nx_pointsC=None, x_stepC=None,

                 x_minQ2=None, x_maxQ2=None, nx_pointsQ2=None, x_stepQ2=None,
                 ny_pointsQ2=None, y_stepQ2=None,

                 hdawg_port: str = '127.0.0.1', hdawg_port1: int = 8004, hdawg_mode: int = 6,
                 hdawg_device: str = 'dev8210',
                 LO_res_port: str = 'TCPIP0::192.168.180.110::inst0::INSTR',
                 LO_res_set_bandwidth: int = 20, LO_res_set_power: int = -10,
                 LO_res_set_averages=1
                 ):

        if (x_min is not None) and (x_max is not None):

            x_minQ1 = x_min if x_minQ1 is None else x_minQ1
            x_minQ2 = x_min if x_minQ2 is None else x_minQ2
            x_minC = x_min if x_minC is None else x_minC

            x_maxQ1 = x_max if x_maxQ1 is None else x_maxQ1
            x_maxQ2 = x_max if x_maxQ2 is None else x_maxQ2
            x_maxC = x_max if x_maxC is None else x_maxC

            nx_pointsQ1 = nx_points if nx_pointsQ1 is None else nx_pointsQ1
            nx_pointsQ2 = nx_points if nx_pointsQ2 is None else nx_pointsQ2
            nx_pointsC = nx_points if nx_pointsC is None else nx_pointsC

            x_stepQ1 = x_step if x_stepQ1 is None else x_stepQ1
            x_stepQ2 = x_step if x_stepQ2 is None else x_stepQ2
            x_stepC = x_step if x_stepC is None else x_stepC

        else:
            pass

        "Flux on Q1 | Coupler | Q2; Readout - Q1"
        self.Q1 = Sts2Q(x_min=x_minQ1, x_max=x_maxQ1, y_min=y_minQ1, y_max=y_maxQ1,
                        nx_points=nx_pointsQ1, x_step=x_stepQ1, y_step=y_stepQ1, ny_points=ny_pointsQ1,
                        target_ch=q1_ch,
                        readout_qubit='Q1', flux_qubit='Q1',
                        hdawg_port=hdawg_port, hdawg_port1=hdawg_port1, hdawg_mode=hdawg_mode,
                        hdawg_device=hdawg_device,
                        LO_res_port=LO_res_port,
                        LO_res_set_bandwidth=LO_res_set_bandwidth, LO_res_set_power=LO_res_set_power,
                        LO_res_set_averages=LO_res_set_averages)

        self.Q1_Coupler = Sts2Q(x_min_coupler=x_minC, x_max_coupler=x_maxC, y_min=y_minQ1, y_max=y_maxQ1,
                                nx_points_coupler=nx_pointsC, x_step_coupler=x_stepC, y_step=y_stepQ1, ny_points=ny_pointsQ1,
                                coupler_ch=coupler_ch,
                                readout_qubit='Q1', flux_qubit='Coupler',
                                hdawg_port=hdawg_port, hdawg_port1=hdawg_port1, hdawg_mode=hdawg_mode,
                                hdawg_device=hdawg_device,
                                LO_res_port=LO_res_port,
                                LO_res_set_bandwidth=LO_res_set_bandwidth, LO_res_set_power=LO_res_set_power,
                                LO_res_set_averages=LO_res_set_averages)

        self.Q1_Q2 = Sts2Q(x_min_control=x_minQ2, x_max_control=x_maxQ2, y_min=y_minQ1, y_max=y_maxQ1,
                           nx_points_control=nx_pointsQ2, x_step_control=x_stepQ2, y_step=y_stepQ1, ny_points=ny_pointsQ1,
                           control_ch=q2_ch,
                           readout_qubit='Q1', flux_qubit='Q2',
                           hdawg_port=hdawg_port, hdawg_port1=hdawg_port1, hdawg_mode=hdawg_mode,
                           hdawg_device=hdawg_device,
                           LO_res_port=LO_res_port,
                           LO_res_set_bandwidth=LO_res_set_bandwidth, LO_res_set_power=LO_res_set_power,
                           LO_res_set_averages=LO_res_set_averages)

        "Flux on Q1 | Coupler | Q2; Readout - Q2"
        self.Q2 = Sts2Q(x_min=x_minQ2, x_max=x_maxQ2, y_min=y_minQ2, y_max=y_maxQ2,
                        nx_points=nx_pointsQ2, x_step=x_stepQ2, y_step=y_stepQ2, ny_points=ny_pointsQ2,
                        target_ch=q2_ch,
                        readout_qubit='Q2', flux_qubit='Q2',
                        hdawg_port=hdawg_port, hdawg_port1=hdawg_port1, hdawg_mode=hdawg_mode,
                        hdawg_device=hdawg_device,
                        LO_res_port=LO_res_port,
                        LO_res_set_bandwidth=LO_res_set_bandwidth, LO_res_set_power=LO_res_set_power,
                        LO_res_set_averages=LO_res_set_averages)

        self.Q2_Coupler = Sts2Q(x_min_coupler=x_minC, x_max_coupler=x_maxC, y_min=y_minQ2, y_max=y_maxQ2,
                                nx_points_coupler=nx_pointsC, x_step_coupler=x_stepC, y_step=y_stepQ2, ny_points=ny_pointsQ2,
                                coupler_ch=coupler_ch,
                                readout_qubit='Q2', flux_qubit='Coupler',
                                hdawg_port=hdawg_port, hdawg_port1=hdawg_port1, hdawg_mode=hdawg_mode,
                                hdawg_device=hdawg_device,
                                LO_res_port=LO_res_port,
                                LO_res_set_bandwidth=LO_res_set_bandwidth, LO_res_set_power=LO_res_set_power,
                                LO_res_set_averages=LO_res_set_averages)

        self.Q2_Q1 = Sts2Q(x_min_control=x_minQ1, x_max_control=x_maxQ1, y_min=y_minQ2, y_max=y_maxQ2,
                           nx_points_control=nx_pointsQ1, x_step_control=x_stepQ1, y_step=y_stepQ2, ny_points=ny_pointsQ2,
                           control_ch=q1_ch,
                           readout_qubit='Q2', flux_qubit='Q1',
                           hdawg_port=hdawg_port, hdawg_port1=hdawg_port1, hdawg_mode=hdawg_mode,
                           hdawg_device=hdawg_device,
                           LO_res_port=LO_res_port,
                           LO_res_set_bandwidth=LO_res_set_bandwidth, LO_res_set_power=LO_res_set_power,
                           LO_res_set_averages=LO_res_set_averages)

        pass

    @property
    def final_matrix(self):

        # First row
        Q1_offset_base = self.Q1.x_offset if self.Q1.finished else None
        a11 = 1 if self.Q1.finished else None

        Q2_offset = self.Q1_Q2.x_offset if self.Q1_Q2.finished else None
        a12 = Q1_offset_base / Q2_offset if self.Q1_Q2.finished and self.Q1.finished else None

        Q1_coupler_offset = self.Q1_Coupler.x_offset if self.Q1_Coupler.finished else None
        a13 = Q1_offset_base / Q1_coupler_offset if self.Q1_Coupler.finished and self.Q1.finished else None

        # Second row
        Q2_offset_base = self.Q2.x_offset if self.Q2.finished else None
        Q1_offset = self.Q2_Q1.x_offset if self.Q2_Q1.finished else None
        a21 = Q2_offset_base / Q1_offset if self.Q2_Q1.finished and self.Q2.finished else None

        a22 = 1 if self.Q2.finished else None

        Q2_coupler_offset = self.Q2_Coupler.x_offset if self.Q2_Coupler.finished else None
        a23 = Q2_offset_base / Q2_coupler_offset if self.Q2_Coupler.finished and self.Q2.finished else None

        # Third row
        a31 = 0
        a32 = 0
        a33 = 1

        matrix = np.array([[a11, a12, a13], [a21, a22, a23], [a31, a32, a33]])
        return np.linalg.inv(matrix)





import zhinst
import zhinst.ziPython
from drivers.M9290A import *
from drivers.N5183B import *
from drivers.znb_besedin import *
from resonator_tools import circlefit
from resonator_tools.circuit import notch_port

from pyquac.fmn_datatools import TwoToneSpectroscopy, timer

# built-in libraries
from time import perf_counter
from typing import Iterable, Union
import time
# installable libraries
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from scipy import stats
import numba as nb


class Tts(TwoToneSpectroscopy):

    def __init__(self,
                 *, x_min, x_max, y_min, y_max,
                 fr_min: float, fr_max: float,
                 hdawg_port: str = '127.0.0.1', hdawg_port1: int = 8004, hdawg_port2: int = 6,
                 LO_port: str = 'TCPIP0::192.168.180.143::hislip0::INSTR',
                 LO_res_port: str = 'TCPIP0::192.168.180.110::inst0::INSTR',
                 hdawg_channel: int = 5, hdawg_setDouble: str = '/dev8210/sigouts/5/offset',
                 LO_set_power: int = 5,
                 LO_res_set_bandwidth: int = 20, LO_res_set_power: int = -10, LO_res_set_nop=101,
                 base_bandwidth=40, LO_res_set_averages=1, LO_res_meas_averages=1, nx_points=None, x_step=None,
                 y_step=None, ny_points=None
                 ):
        """
        Class provides methods for working with live data for Two Tone Spectroscopy
        :param x_min: x minimum value (int | float)
        :param x_max: x maximum value (int | float)
        :param nx_points: x count value (int)
        :param y_min: y minimum value (int | float)
        :param y_max: y maximum value (int | float)
        :param ny_points: y count value (int)
        :param fr_min: min frequency for find resonator
        :param fr_max: max frequency for find resonator
        :param hdawg_port: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_port2)
        :param hdawg_port1: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_port2)
        :param hdawg_port2: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_port2)
        :param LO_port: qubit LO = N5183B('N5183B', LO_port)
        :param LO_res_port: resonator LO_res = Znb(LO_res_port)
        :param hdawg_channel: hdawg.setInt('/dev8210/sigouts/' + str(channel) + '/on', 1)
        :param LO_set_power: base LO power (default 5)
        :param LO_res_set_bandwidth: bandwidth during resonator tuning
        :param LO_res_set_power: base LO_res power (default -10)
        :param base_bandwidth: (int) bandwidth during mesurments
        """
        super().__init__(x_min=x_min, x_max=x_max, nx_points=nx_points, y_min=y_min, y_max=y_max, ny_points=ny_points,
                         x_step=x_step, y_step=y_step)
        self.fr_min = fr_min
        self.fr_max = fr_max

        # HDAWG init
        self.hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_port2)
        self.hdawg_setDouble = hdawg_setDouble
        # open HDAWG ZI
        hdawgModule = self.hdawg.awgModule()
        channel = hdawg_channel
        self.hdawg.setInt('/dev8210/sigouts/' + str(channel) + '/on', 1)

        # freq generator init
        self.LO = N5183B('N5183B', LO_port)  # cubit
        self.LO_res = Znb(LO_res_port)  # resonator

        self.LO_set_power = LO_set_power

        self.LO_res_set_nop = LO_res_set_nop
        self.LO_res_set_bandwidth = LO_res_set_bandwidth
        self.LO_res_set_power = LO_res_set_power
        self.base_bandwidth = base_bandwidth
        self.LO_res_set_averages = LO_res_set_averages
        self.LO_res.set_averages(LO_res_set_averages)
        self.LO_res_meas_averages = LO_res_meas_averages

        pass

    @property
    def __find_resonator(self):

        self.LO.set_status(0)

        # prior bandwidth
        bandwidth = self.LO_res.get_bandwidth()
        xlim = self.LO_res.get_freq_limits()

        self.LO_res.set_bandwidth(self.LO_res_set_bandwidth)
        self.LO_res.set_nop(self.LO_res_set_nop)
        self.LO_res.set_freq_limits(self.fr_min, self.fr_max)
        self.LO_res.set_averages(self.LO_res_set_averages)

        # measure S21
        freqs = self.LO_res.get_freqpoints()
        notch = notch_port(freqs, self.LO_res.measure()['S-parameter'])
        notch.autofit(electric_delay=60e-9)
        result = round(notch.fitresults['fr'])

        # Resetting to the next round of measurements
        self.LO_res.set_bandwidth(bandwidth)
        self.LO_res.set_freq_limits(*xlim)
        self.LO_res.set_nop(1)
        self.LO.set_status(1)

        return result

    def run_measurements(self, *, x_key: Union[float, int, Iterable] = None, y_key: Union[float, int, Iterable] = None,
                         x_min=None, x_max=None, y_min=None, y_max=None,
                         sleep=0.007, timeout=None):

        self.iter_setup(x_key=x_key, y_key=y_key,
                        x_min=x_min, x_max=x_max, y_min=y_min, y_max=y_max)

        # Set power
        self.LO.set_power(self.LO_set_power)
        self.LO_res.set_power(self.LO_res_set_power)
        # base bandwidth
        self.LO_res.set_bandwidth(self.base_bandwidth)
        try:
            for i in range(len(self.load)):
                if (i == 0) or (self.load[i] != self.load[i - 1]):
                    self.LO_res.set_center(float(self.__find_resonator))
                    if timeout is not None:
                        timer.sleep(timeout)
                    else:
                        pass
                # measurement averages
                self.LO_res.set_averages(self.LO_res_meas_averages)

                self.hdawg.setDouble(self.hdawg_setDouble, self.load[i])  # Current write
                self.LO.set_frequency(self.frequency[i])  # Frequency write

                result = self.LO_res.measure()['S-parameter']

                self.write(x=self.load[i],
                           y=self.frequency[i],
                           heat=20 * np.log10(abs(result)[0])
                           )
                timer.sleep(sleep)

            self.LO.set_status(0)
        except KeyboardInterrupt:
            pass


class Sts(TwoToneSpectroscopy):

    def __init__(self,
                 *, x_min, x_max, y_min, y_max,
                 hdawg_port: str = '127.0.0.1', hdawg_port1: int = 8004, hdawg_port2: int = 6,
                 LO_res_port: str = 'TCPIP0::192.168.180.110::inst0::INSTR',
                 hdawg_channel: int = 5, hdawg_setDouble: str = '/dev8210/sigouts/5/offset',
                 LO_res_set_bandwidth: int = 20, LO_res_set_power: int = -10,
                 LO_res_set_averages=1,
                 nx_points=None, x_step=None, y_step=None, ny_points=None
                 ):
        """
        Class provides methods for working with live data for Two Tone Spectroscopy
        :param x_min: x minimum value (int | float)
        :param x_max: x maximum value (int | float)
        :param nx_points: x count value (int)
        :param y_min: y minimum value (int | float)
        :param y_max: y maximum value (int | float)
        :param ny_points: y count value (int)

        :param hdawg_port: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_port2)
        :param hdawg_port1: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_port2)
        :param hdawg_port2: hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_port2)
        :param LO_res_port: resonator LO_res = Znb(LO_res_port)
        :param hdawg_channel: hdawg.setInt('/dev8210/sigouts/' + str(channel) + '/on', 1)
        :param LO_res_set_bandwidth: bandwidth during resonator tuning
        :param LO_res_set_power: base LO_res power (default -10)
        """
        super().__init__(x_min=x_min, x_max=x_max, nx_points=nx_points, y_min=y_min, y_max=y_max, ny_points=ny_points,
                         x_step=x_step, y_step=y_step)

        # HDAWG init
        self.hdawg = zhinst.ziPython.ziDAQServer(hdawg_port, hdawg_port1, hdawg_port2)
        self.hdawg_setDouble = hdawg_setDouble
        # open HDAWG ZI
        hdawgModule = self.hdawg.awgModule()
        channel = hdawg_channel
        self.hdawg.setInt('/dev8210/sigouts/' + str(channel) + '/on', 1)

        # freq generator init
        self.LO_res = Znb(LO_res_port)  # resonator

        self.LO_res_set_nop = LO_res_set_nop
        self.LO_res_set_bandwidth = LO_res_set_bandwidth
        self.LO_res_set_power = LO_res_set_power

        self.LO_res_set_averages = LO_res_set_averages
        self.LO_res.set_averages(LO_res_set_averages)

        pass

    def run_measurements(self, *, sleep=0.0007):

        self.iter_setup(x_key=None, y_key=None,
                        x_min=None, x_max=None, y_min=None, y_max=None)

        # Set power
        self.LO_res.set_power(self.LO_res_set_power)

        # Set parameters
        self.LO_res.set_bandwidth(self.LO_res_set_bandwidth)
        self.LO_res.set_nop(self.ny_points)
        self.LO_res.set_freq_limits(self.y_min, self.y_max)
        freqs = self.LO_res.get_freqpoints()

        freq_len = len(self.y_list)

        try:
            for i in range(len(self.load)):
                if (i == 0) or (self.load[i] != self.load[i - 1]):
                    self.hdawg.setDouble(self.hdawg_setDouble, self.load[i])  # Current write

                    self.LO_res.set_averages(self.LO_res_set_averages)

                    # measure S21
                    notch = notch_port(freqs, self.LO_res.measure()['S-parameter'])
                    notch.autofit(electric_delay=60e-9)
                    result = notch.fitresults['fr']

                    for j in range(freq_len):
                        self.write(x=self.load[i],
                                   y=self.y_list[j],
                                   heat=result[j]
                                   )

                timer.sleep(sleep)

        except KeyboardInterrupt:
            self.drop(x=self.x_raw[-1])
            pass


# built-in libraries
from time import perf_counter
from typing import Iterable, Union
import time
# installable libraries
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from scipy import stats
import numba as nb


@nb.jit(nopython=True)
def _complicated(raw_array_x, raw_array_y, x_min_, x_step_, y_min_, y_step_,
                 len_y_):
    ind_array = np.zeros(len(raw_array_x))

    if len(ind_array) >= 2:
        for k in range(len(raw_array_x)):
            x_ind = np.around((raw_array_x[k] - x_min_) / x_step_)
            y_ind = np.around((raw_array_y[k] - y_min_) / y_step_)

            ind_array[k] = x_ind * len_y_ + y_ind
            pass

        return ind_array
    else:
        pass


class mrange:

    @classmethod
    def cust_range(cls, *args, rtol=1e-05, atol=1e-08, include=[True, False]):
        """
        Combines numpy.arange and numpy.isclose to mimic
        open, half-open and closed intervals.
        Avoids also floating point rounding errors as with
        numpy.arange(1, 1.3, 0.1)
        array([1. , 1.1, 1.2, 1.3])

        args: [start, ]stop, [step, ]
            as in numpy.arange
        rtol, atol: floats
            floating point tolerance as in numpy.isclose
        include: boolean list-like, length 2
            if start and end point are included
        """
        # process arguments
        if len(args) == 1:
            start = 0
            stop = args[0]
            step = 1
        elif len(args) == 2:
            start, stop = args
            step = 1
        else:
            assert len(args) == 3
            start, stop, step = tuple(args)

        # determine number of segments
        n = (stop - start) / step + 1

        # do rounding for n
        if np.isclose(n, np.round(n), rtol=rtol, atol=atol):
            n = np.round(n)

        # correct for start/end is exluded
        if not include[0]:
            n -= 1
            start += step
        if not include[1]:
            n -= 1
            stop -= step

        return np.linspace(start, stop, int(n))

    @classmethod
    def crange(cls, *args, **kwargs):
        return mrange.cust_range(*args, **kwargs, include=[True, True])

    @classmethod
    def orange(cls, *args, **kwargs):
        return mrange.cust_range(*args, **kwargs, include=[True, False])


class timer:
    """
    timer class provides accurate variation of time.sleep() method. Example:
    timer.sleep(1) #  sleeps for 1. second

    Why to use this? If you would use time.sleep() for 1 iteration, it wouldn't bring you problem, but if your script
    requires a loop with time.sleep() inside, windows OS will ruin your script execution. To explain why this
    happened, Windows OS has its default timer resolution set to 15.625 ms or 64 Hz, which is decently enough for
    most of the applications. However, for applications that need very short sampling rate or time delay, then 15.625
    ms is not sufficient.
    """

    @staticmethod
    def sleep(sec: float):
        deadline = perf_counter() + sec
        while perf_counter() < deadline:
            pass


class TwoToneSpectroscopy:

    def __init__(self,
                 *, x_min, x_max, y_min, y_max,
                 x_step=None, y_step=None, nx_points=None, ny_points=None):
        """
        Class provides methods for working with live data for Two Tone Spectroscopy
        :param x_min: x minimum value (int | float)
        :param x_max: x maximum value (int | float)
        :param x_step: x step value (int)
        :param y_min: y minimum value (int | float)
        :param y_max: y maximum value (int | float)
        :param y_step: y step value (int)
        """
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max

        if nx_points is not None:
            _, self.x_step = np.linspace(x_min, x_max, nx_points, retstep=True)
            self.x_step = round(self.x_step, 10)
        else:
            self.x_step = float(x_step)

        if ny_points is not None:
            _, self.y_step = np.linspace(y_min, y_max, ny_points, retstep=True)
            self.y_step = round(self.y_step, 2)
        else:
            self.y_step = float(y_step)

        self.load = None
        self.frequency = None

        self.__x_step_DP = len(str(self.x_step).split(".")[1])
        self.__y_step_DP = len(str(self.y_step).split(".")[1])
        self.__max_step_DP = max([self.__x_step_DP, self.__y_step_DP])

        self.x_list = mrange.orange(self.x_min, self.x_max + self.x_step,
                                    self.x_step)
        self.y_list = mrange.orange(self.y_min, self.y_max + self.y_step,
                                    self.y_step)

        self.x_list = np.around(self.x_list, decimals=self.__x_step_DP)
        self.y_list = np.around(self.y_list, decimals=self.__y_step_DP)

        self.x_raw = []
        self.y_raw = []
        self.heat_raw = []

        self.heat_container = np.zeros(len(self.x_list) * len(self.y_list))
        self.heat_container[:] = np.nan

        self.heat_container1 = np.zeros(len(self.x_list) * len(self.y_list))
        self.heat_container1[:] = np.nan

        self.len_y = len(self.y_list)

        self.x_container = np.zeros(len(self.x_list) * len(self.y_list))
        self.y_container = np.zeros(len(self.x_list) * len(self.y_list))

        self.nx_steps = len(self.x_list)
        self.ny_steps = len(self.y_list)

        self.heat_list = np.empty((self.ny_steps, self.nx_steps))
        self.heat_list[:, :] = np.nan

        self.heat_samples = np.array([])
        self.heat_mins = np.array([])
        self.heat_maxs = np.array([])

        self.glob_h_mi = 0
        self.glob_h_ma = 0
        self.glob_samp = 0

        self.x_1d = np.repeat(self.x_list, len(self.y_list))
        self.y_1d = np.tile(self.y_list, len(self.x_list))

        # service variables
        self.__n_steps = int(0.04 * len(self.y_list))

        x1 = 2000
        y1 = 70e3

        x2 = 3000
        y2 = 350e3

        k = (y1 - y2) / (x1 - x2)
        self.timing = np.arange(1000, 40000, 100)

        b = y2 - k * x2

        self.count_of_elements = k * self.timing + b

        self.x_for_approximate_idxs = []

        self.active_state = True

    def iter_setup(self, *, x_key: Union[float, int, Iterable] = None, y_key: Union[float, int, Iterable] = None,
                   x_min=None, x_max=None, y_min=None, y_max=None):
        """
        measurement setup. if all optional params are None then setup self.load and self.frequency for measuring all data
        :param x_key: x value(s) for measurement
        :param y_key: y value(s) for measurement
        :param x_min: x min value of measurement
        :param x_max: x max value of measurement
        :param y_min: y min value of measurement
        :param y_max: y max value of measurement
        :return:
        """
        if x_key is not None:
            x_key = self.__config_closest_values(x_key, self.x_list)

        if y_key is not None:
            if np.array(y_key).dtype == object:
                pass
            else:
                y_key = self.__config_closest_values(y_key, self.y_list)

        "set x_min x_max arrays"
        if (y_min != None) | (y_max != None) | (x_min != None) | (x_max != None):
            x_key = self.__config_arrays_from(min_value=x_min, max_value=x_max, step=self.x_step,
                                              array=self.x_list, array_set_input_in_func=x_key, column='x_value')

            y_key = self.__config_arrays_from(min_value=y_min, max_value=y_max, step=self.y_step,
                                              array=self.y_list, array_set_input_in_func=y_key, column='y_value')

        if x_key is not None:
            x_key = x_key if isinstance(x_key, Iterable) else [x_key, ]

        x = np.array(x_key) if x_key is not None else np.array(self.x_list)

        """config frequencies"""
        if y_key is not None:
            y_key = y_key if isinstance(y_key, Iterable) else [y_key, ]
            if np.array(y_key).dtype != object:
                y_key = np.tile(y_key, [len(x), 1])
            elif np.array(y_key).dtype == object and x_key is not None:
                ind_list = []
                for x_k in x_key:
                    ind_list.append(list(self.x_list).index(x_k))
                y_key = y_key[ind_list]

            else:
                pass

            if np.array(y_key).dtype == object:
                freqs = []
                for i in range(len(y_key)):
                    freqs += list(y_key[i])
                freqs = np.array(freqs)

            else:
                freqs = y_key.ravel()


        else:
            freqs = np.tile(self.y_list, len(x))
            pass

        """config currents"""
        if y_key is not None:
            if np.array(y_key).dtype == object:
                current_encapsulated = []
                for i, y_i in enumerate(y_key):
                    current_encapsulated.append([x[i], ] * len(y_i))
                # currents = np.array(current_encapsulated).ravel()

                if np.array(y_key).dtype == object:
                    currents = []
                    for i in range(len(current_encapsulated)):
                        currents += list(current_encapsulated[i])
                    currents = np.array(currents)

                else:
                    currents = np.array(current_encapsulated).ravel()

            else:
                current_encapsulated = []
                for i, y_i in enumerate(y_key):
                    current_encapsulated.append([x[i], ] * len(y_i))
                currents = np.array(current_encapsulated).ravel()
                pass

        else:
            currents = np.repeat(x, len(self.y_list))

        "collect all together"
        temp_df = pd.DataFrame({'x_value': currents, 'y_value': freqs})
        raw_frame_without_nans = self.raw_frame.dropna(axis=0)

        index1 = pd.MultiIndex.from_arrays([temp_df[col] for col in ['x_value', 'y_value']])
        index2 = pd.MultiIndex.from_arrays([raw_frame_without_nans[col] for col in ['x_value', 'y_value']])
        temp_df = temp_df.loc[~index1.isin(index2)]

        self.load = temp_df['x_value'].values
        self.frequency = temp_df['y_value'].values

        pass

    def write(self, *, x: Union[float, int] = None, y: Union[float, int] = None,
              heat: Union[float, int] = None):
        """
        writes load, frequency ans response to class entity
        :param x: load value
        :param y: frequency value
        :param z: response value
        """
        self.x_raw.append(x)  # Current write to DF
        self.y_raw.append(y)  # Frequency write to DF
        self.heat_raw.append(heat)  # Response write to DF
        pass

    def check_stop_on_iter(self, i):
        if len(self.x_raw) == len(self.y_raw) == len(self.heat_raw):
            pass
        else:
            max_l = max([len(self.x_raw), len(self.y_raw), len(self.heat_raw)])
            i = i + 1 if i < len(self.frequency) else i
            if len(self.y_raw) < max_l: self.y_raw.append(self.frequency[i])
            if len(self.heat_raw) < max_l: self.heat_raw.append(np.nan)

    @property
    def raw_frame(self):
        """
        generates raw Data Frame with columns [x_value, y_value, heat_value]
        :return: dataframe
        """
        if len(self.x_raw) == len(self.y_raw) == len(self.heat_raw):
            return pd.DataFrame({'x_value': self.x_raw, 'y_value': self.y_raw, 'heat_value': self.heat_raw})
        else:
            pass

    @property
    def raw_array(self):
        """
        generates raw 2-d numpy array np.array([y_value, x_value])
        :return:
        """
        if len(self.x_raw) == len(self.y_raw) == len(self.heat_raw):
            return np.array([self.x_raw, self.y_raw, self.heat_raw])
        else:
            pass

    def get_result(self, *, imshow: bool = False) -> pd.DataFrame:
        """
        return resulting Data Frame
        :param imshow: (optional) if True then result returns dataframe for matplotlib.pyplot.imshow()
        :return: Pandas Data Frame. column.names: ['currents', 'frequencies', 'response']
        """
        l_h = len(self.heat_container[:len(self.heat_raw)])
        l_x = len(self.x_container[:len(self.x_raw)])
        l_y = len(self.y_container[:len(self.y_raw)])

        if l_h == l_x == l_y:
            self.heat_container[:len(self.heat_raw)] = self.heat_raw
            self.x_container[:len(self.x_raw)] = self.x_raw
            self.y_container[:len(self.y_raw)] = self.y_raw
        else:
            pass

        for i in range(len(self.x_raw)):
            self.heat_list[round((self.y_container[i] - self.y_min) / self.y_step),
                           round((self.x_container[i] - self.x_min) / self.x_step)] = self.heat_container[i]

        heat_1d = self.heat_list.ravel(order='F')

        if not imshow:
            df = pd.DataFrame({'currents': self.x_1d, 'frequencies': self.y_1d, 'response': heat_1d})
        else:
            df = pd.DataFrame(data=self.heat_list, columns=self.x_list, index=self.y_list)

        return df

    def get_raw_result(self):
        """
        generates raw Data Frame with columns [x_value, y_value, response]
        :return: Pandas Data Frame
        """
        x = []
        y = []
        heat = []

        uniq_x = np.unique(self.raw_frame.x_value.values)

        for xx in uniq_x:
            y_arr = self.raw_frame[self.raw_frame.x_value == xx].y_value.values
            heat_arr = self.raw_frame[self.raw_frame.x_value == xx].heat_value.values

            x.append(xx)
            y.append(y_arr)
            heat.append(heat_arr)

        return pd.DataFrame({'x_value': x, 'y_value': y, 'response': heat})

    @property
    def non_njit_result(self):
        """
        generates resulting 2-d array
        :return: numpy.ndarray[y_values, x_values]
        """

        if len(self.x_raw) >= 2:
            array = np.array(self.raw_array)
            z_val = array[2, :]
            x_val = array[0, :]
            y_val = array[1, :]

            for i in range(len(z_val)):
                self.heat_list[round((y_val[i] - self.y_min) / self.y_step),
                               round((x_val[i] - self.x_min) / self.x_step)] = z_val[i]

            return self.heat_list

        else:
            pass

    @property
    def njit_result(self):
        """
        generates CPU accelerated resulting 2-d array
        :return: numpy.ndarray[y_values, x_values]
        """
        # array = np.copy(self.raw_array)
        array_to_njit = self.raw_array
        z_val = array_to_njit[2, :]
        x_val = array_to_njit[0, :]
        y_val = array_to_njit[1, :]

        ind_array = _complicated(x_val, y_val,
                                 self.x_min, self.x_step, self.y_min, self.y_step,
                                 self.len_y)
        if len(self.x_raw) >= 2:
            array_to_process = np.copy(self.heat_container1)
            array_to_process[ind_array.astype(int)] = z_val

            return array_to_process
        else:
            pass

    def approximate(self, resolving_zone: float = 0.1, *, fillna: bool = False, info=False,
                    plot: bool = False, heat_sample=None, heat_sampe_tol=0.5, x_key=None):
        """
        return dict object with keys:
        1) y_key - an object containing y values for each x of self.x_list
        2) mask - mask of values that would be measured. it can be plotted with plotly's Heatmap
        3) imshow_mask - mask of values that would be measured. it can be plotted with plotly's Heatmap
        4) poly_line - approximated np.poly1d() curve
        :param resolving_zone: [0:1) - resolving zone for plot
        :param fillna: (optional) if True then fill plot with response minimum value
        :param info: (optional) set True if you would like to get information of approximation process
        :param plot: (optional) set True if you would like to get density plots of outliers on measured x values
        :return: dictionary. keys: ['y_key', 'mask', 'imshow_mask', 'poly_line']
        """

        if x_key is not None:
            x_key = self.__config_closest_values(x_key, self.x_list)
            X = self.raw_frame[self.raw_frame.x_value.isin(x_key)].copy()
            X = X.reset_index(drop=True)
            x_set = x_key
        else:
            X = self.raw_frame.copy()
            """rows that we are looking for (every x_value)"""
            x_set = np.unique(X['x_value'].values)  # get an array of unique values of x

        if heat_sample is None:
            glob_samp, glob_h_mi, glob_h_ma = self.__define_heat_sample_on(x_set, info=info, plot=plot, q_max=90,
                                                                           q_min=10)

            self.glob_samp = glob_samp
            self.glob_h_mi = glob_h_mi
            self.glob_h_ma = glob_h_ma

            for xxx in x_set:
                heat_sample, heat_min, heat_max = self.__define_heat_sample_on(xxx, info=info, plot=plot)

                self.heat_samples = np.append(self.heat_samples, heat_sample)
                self.heat_mins = np.append(self.heat_mins, heat_min)
                self.heat_maxs = np.append(self.heat_maxs, heat_max)

            self.__change_outlier_heatsamples()  # change bad heat samples
            pass

        else:
            self.glob_samp = heat_sample
            self.glob_h_mi = heat_sample - heat_sampe_tol
            self.glob_h_ma = heat_sample + heat_sampe_tol

            for _ in x_set:
                self.heat_samples = np.append(self.heat_samples, self.glob_samp)
                self.heat_mins = np.append(self.heat_mins, self.glob_h_mi)
                self.heat_maxs = np.append(self.heat_maxs, self.glob_h_ma)

        tuple_list = ()
        ip = 0
        for xx in x_set:

            heat_sample, heat_min, heat_max = self.heat_samples[ip], self.heat_mins[ip], self.heat_maxs[ip]
            ip += 1
            # bad_y = self.__bad_outliers_on(xx)

            y_min, y_max = self.__find_min_max_y_on_(xx, heat_sample)

            if len(X[X.x_value == xx].y_value) <= 0.7 * len(self.y_list):
                temp_y_idx = (X[X.x_value == xx].heat_value - heat_sample).abs().idxmin()
            else:
                temp_y_idx = (X[(X.x_value == xx) & (X.y_value >= y_min) & (X.y_value <= y_max)
                                ].heat_value - heat_sample).abs().idxmin()

            temp_max_row = tuple(X.iloc[temp_y_idx])

            tuple_list += (temp_max_row,)

        # clear arrays
        self.heat_samples = np.array([])
        self.heat_mins = np.array([])
        self.heat_maxs = np.array([])

        # rotating array
        tuple_of_max_z_values = np.array(tuple_list).T

        poly = np.poly1d(np.polyfit(x=tuple_of_max_z_values[0], y=tuple_of_max_z_values[1], deg=2))

        print('x', tuple_of_max_z_values[0])
        print('y', tuple_of_max_z_values[1])

        """getting freq values for approximation"""
        y_for_approximate = []
        self.x_for_approximate_idxs = []
        x_for_approximate_ind = 0
        for value in poly(self.x_list):
            y_for_approximate.append(self.__find_nearest(value))
            if (value <= self.y_min) or (value >= self.y_max):
                self.x_for_approximate_idxs.append(x_for_approximate_ind)
            x_for_approximate_ind += 1

        """masking"""

        min_heat_sample = X.heat_value.min() if fillna is True else np.nan
        max_heat_sample = tuple_of_max_z_values[2].mean()

        get_result_df = self.get_result()
        get_result_df.loc[:, 'response'] = min_heat_sample

        for i in range(len(y_for_approximate)):
            get_result_mask = (get_result_df['currents'] == self.x_list[i]) & (
                    get_result_df['frequencies'] == y_for_approximate[i])
            get_result_df.loc[get_result_mask, 'response'] = max_heat_sample

        """approximation"""
        y_keys = []

        for xx in self.x_list:
            idx = get_result_df[get_result_df.currents == xx].loc[get_result_df.response == max_heat_sample].index[0]

            count_of_resolve_idx = len(self.y_list) * resolving_zone
            get_result_df.iloc[idx + 1:idx + round(count_of_resolve_idx / 2), 2] = max_heat_sample / 2

            get_result_df.iloc[idx - round(count_of_resolve_idx / 2):idx, 2] = max_heat_sample / 2
            y_keys.append(
                get_result_df.iloc[idx - int(count_of_resolve_idx / 2):idx + int(count_of_resolve_idx / 2),
                1].to_list())

            self.approximation_y_keys = np.array(y_keys, dtype=object)

        """checking for zero-length list. and fix it. what did I do here???????"""
        for i in range(len(y_keys)):
            if len(y_keys[i]) == 0:
                if i == len(y_keys) - 1:
                    y_keys[i] = list(y_keys[i - 1])
                else:
                    y_keys[i] = list(y_keys[i + 1])

        """config returning values"""
        """imshow"""
        heat_list = []
        for xx in self.x_list:
            heat_list.append(get_result_df[get_result_df.currents == xx].loc[:, 'response'].values)
        get_imshow_result_df = pd.DataFrame(data=np.array(heat_list).T, columns=self.x_list, index=self.y_list)

        "deleting bad x approximation"
        # y_keys = np.delete(y_keys, self.x_for_approximate_idxs)
        x_keys = np.delete(self.x_list, self.x_for_approximate_idxs)

        return dict(y_key=np.array(y_keys, dtype=object),
                    x_key=x_keys,
                    mask=get_result_df,
                    imshow_mask=get_imshow_result_df,
                    poly_line={'x': self.x_list, 'y': poly(self.x_list)})

    def fit_curve(self, x_key=None):

        X = self.raw_frame.copy()

        """rows that we are looking for (every x_value)"""
        x_set = np.unique(X['x_value'].values)  # get an array of unique values of x

        glob_samp, glob_h_mi, glob_h_ma = self.__define_heat_sample_on(x_set, info=False, plot=False, q_max=95,
                                                                       q_min=5)

        if x_key is not None:
            approx = self.approximate(x_key=x_key, heat_sample=glob_samp, heat_sampe_tol=abs(glob_h_ma - glob_h_mi) / 2)
            y_approx = approx['poly_line']['y']

        tuple_list = ()
        ip = 0
        for xx in x_set:

            y_min, y_max = self.__find_min_max_y_on_(xx, glob_samp)

            if x_key is not None:
                "block to prevent jump in y axis"
                y_mid = (y_max + y_min) / 2
                if abs(y_mid - y_approx[ip]) > (self.y_max - self.y_min) * 0.07:
                    y_min = y_approx[ip] - self.__n_steps * self.y_step
                    y_max = y_approx[ip] + self.__n_steps * self.y_step
                pass

            ip += 1

            if len(self.raw_frame[X.x_value == xx].y_value) <= 0.7 * len(self.y_list):
                temp_y_idx = (X[X.x_value == xx].heat_value - glob_samp).abs().idxmin()
            else:
                temp_y_idx = (X[(X.x_value == xx) & (X.y_value >= y_min) & (X.y_value <= y_max)
                                ].heat_value - glob_samp).abs().idxmin()

            temp_max_row = tuple(X.iloc[temp_y_idx])

            tuple_list += (temp_max_row,)

        # rotating array
        tuple_of_max_z_values = np.array(tuple_list).T

        return dict(x=tuple_of_max_z_values[0],
                    y=tuple_of_max_z_values[1])

    def check_heat_samp(self, q_max=90, q_min=10):
        X = self.raw_frame.copy()

        """rows that we are looking for (every x_value)"""
        x_set = np.unique(X['x_value'].values)  # get an array of unique values of x

        glob_samp, glob_h_mi, glob_h_ma = self.__define_heat_sample_on(x_set, info=False, plot=False, q_max=q_max,
                                                                       q_min=q_min)
        return glob_samp

    def clean_up(self):
        """
        cleans data after approximation
        """
        i = 0
        tolerance = 1e-3
        for x in np.array(self.x_list, dtype=float):
            # array1 = np.array(self.raw_frame['y_value'].to_numpy(), dtype=float)
            array2 = np.array(self.approximation_y_keys[i], dtype=float)
            mask_arr = ~np.isclose(self.raw_frame['y_value'].values[:, None], array2, atol=.1).any(axis=1)
            idx = self.raw_frame[(mask_arr) & (self.raw_frame['x_value'] == x)].index
            #             mask_array = np.zeros(len(array1), dtype=bool)
            #             for value in array2:
            #                 for j in range(len(array1)):
            #                     if abs(value - array1[j]) < tolerance:
            #                         mask_array[j] = True

            # self.raw_frame[~np.isclose(self.raw_frame['y_value'].values[:, None], array2, atol=.001).any(axis=1)]
            # mask = ~(np.isin(array1,array2))

            self.x_raw = np.delete(self.x_raw, idx)
            self.y_raw = np.delete(self.y_raw, idx)
            self.heat_raw = np.delete(self.heat_raw, idx)
            self.heat_list[:, i] = self.heat_list[:, i] * np.nan
            i += 1

        self.x_raw = list(self.x_raw)
        self.y_raw = list(self.y_raw)
        self.heat_raw = list(self.heat_raw)

        self.heat_container[:] = np.nan
        self.x_container[:] = 0
        self.y_container[:] = 0

    def cls(self):
        """
        clean all heatmap
        """
        self.x_raw = []
        self.y_raw = []
        self.heat_raw = []

        self.heat_list = np.empty((self.ny_steps, self.nx_steps))
        self.heat_list[:, :] = np.nan

        self.heat_container[:] = np.nan
        self.x_container[:] = 0
        self.y_container[:] = 0

    def drop(self, x: Union[float, int, Iterable] = None, y: Union[float, int, Iterable] = None,
             x_min=None, x_max=None, y_min=None, y_max=None):
        """
        drops specific values (x, y)
        :param x: x value(s) [float, int, Iterable]
        :param y: x value(s) [float, int, Iterable]
        :param x_min: minimum x value [float, int]
        :param x_max: maximum x value [float, int]
        :param y_min: minimum y value [float, int]
        :param y_max: maximum y value [float, int]
        """
        """if x list is given then find nearest values in self.x_list"""
        if x is not None:
            x = self.__config_closest_values(x, self.raw_array[0, :])

        """if y list is given then find nearest values in self.y_list"""
        if y is not None:
            y = self.__config_closest_values(y, self.raw_array[1, :])

        "set min max arrays"
        if (y_min != None) | (y_max != None) | (x_min != None) | (x_max != None):
            x = self.__config_arrays_from(min_value=x_min, max_value=x_max, step=self.x_step,
                                          array=self.raw_array[0, :], array_set_input_in_func=x, column='x_value')

            y = self.__config_arrays_from(min_value=y_min, max_value=y_max, step=self.y_step,
                                          array=self.raw_array[1, :], array_set_input_in_func=y, column='y_value')

        if (x is not None) and (y is None):
            self.__drop_the('x_value', x)
            pass

        if (y is not None) and (x is None):
            self.__drop_the('y_value', y)
            pass

        if (x is not None) and (y is not None):
            self.__drop_the_cols(x, y)
            pass

    def __drop_the(self, column: str, value_s):

        decimals = self.__x_step_DP if column == 'x_value' else self.__y_step_DP

        value_s = list(value_s) if isinstance(value_s, Iterable) else [value_s, ]
        value_s = np.around(value_s, decimals=decimals)
        idx = self.raw_frame[self.raw_frame[column].isin(value_s)].index

        self.x_raw = list(np.delete(self.x_raw, idx))
        self.y_raw = list(np.delete(self.y_raw, idx))
        self.heat_raw = list(np.delete(self.heat_raw, idx))

        # for i in idx:
        #     self.heat_raw[i] = np.nan

        self.heat_container[:] = np.nan
        self.x_container[:] = 0
        self.y_container[:] = 0

        column_var = list(self.x_list) if column == 'x_value' else list(self.y_list)

        col_idx_list = [list(column_var).index(value_s[i]) for i in range(len(value_s))]

        if column == 'x_value':
            for i in col_idx_list:
                self.heat_list[:, i] = self.heat_list[:, i] * np.nan
        else:
            for i in col_idx_list:
                self.heat_list[i] = self.heat_list[i] * np.nan

        pass

    def __drop_the_cols(self, x_values, y_values):

        x_decimals, y_decimals = self.__x_step_DP, self.__y_step_DP

        # x mask
        x_value_s = list(x_values) if isinstance(x_values, Iterable) else [x_values, ]
        x_value_s = np.around(x_value_s, decimals=x_decimals)
        x_mask = np.logical_or.reduce(np.isclose(self.raw_frame['x_value'].to_numpy()[None, :], x_value_s[:, None]))

        # y mask
        y_value_s = list(y_values) if isinstance(y_values, Iterable) else [y_values, ]
        y_value_s = np.around(y_value_s, decimals=y_decimals)
        y_mask = np.logical_or.reduce(np.isclose(self.raw_frame['y_value'].to_numpy()[None, :], y_value_s[:, None]))

        # idx
        msk = pd.DataFrame({'y': y_mask, 'x': x_mask}).all(axis=1)
        idx = self.raw_frame[msk].index

        self.x_raw = list(np.delete(self.x_raw, idx))
        self.y_raw = list(np.delete(self.y_raw, idx))
        self.heat_raw = list(np.delete(self.heat_raw, idx))

        self.heat_container[:] = np.nan
        self.x_container[:] = 0
        self.y_container[:] = 0

        x_col_idx_list = [list(self.x_list).index(x_value_s[i]) for i in range(len(x_value_s))]
        y_col_idx_list = [list(self.y_list).index(y_value_s[i]) for i in range(len(y_value_s))]

        for i in x_col_idx_list:
            for j in y_col_idx_list:
                self.heat_list[j, i] = self.heat_list[j, i] * np.nan
        pass

    def __config_closest_values(self, input_value, base_array):

        """if x list is given then find nearest values in self.x_list"""

        if isinstance(input_value, float) or isinstance(input_value, int):
            input_value = self.__find_nearest_universal(base_array, input_value)
        else:
            input_value_temp = []
            for value in input_value:
                input_value_temp.append(self.__find_nearest_universal(base_array, value))
            input_value = np.array(input_value_temp)
        return input_value

    def __config_arrays_from(self, min_value, max_value, step, array, array_set_input_in_func, column='x_value'):

        """configurate arrays from given values"""

        if (min_value is not None) and (max_value is not None):
            min_value = self.__find_nearest_universal(array, min_value)
            max_value = self.__find_nearest_universal(array, max_value)
            array = mrange.orange(min_value, max_value + step, step)
            return array

        elif (min_value is not None) and (max_value is None):
            if column == 'x_value':
                max_v = self.x_max
                step_v = self.x_step
            else:
                max_v = self.y_max
                step_v = self.y_step

            min_value = self.__find_nearest_universal(array, min_value)
            array = mrange.orange(min_value, max_v + step_v, step_v)
            return array
        elif (min_value is None) and (max_value is not None):
            if column == 'x_value':
                min_v = self.x_min
                step_v = self.x_step
            else:
                min_v = self.y_min
                step_v = self.y_step

            max_value = self.__find_nearest_universal(array, max_value)
            array = mrange.orange(min_v, max_value + step_v, step_v)
            return array

        else:
            return array_set_input_in_func

    def __define_heat_sample_on(self, xx, info, plot, q_max=90, q_min=10):

        xx = xx if isinstance(xx, Iterable) else [xx, ]

        """initial params"""
        x_lw = np.array([0, 0])
        y_lw = np.array([0, 0])

        x_uw = np.array([0, 0])
        y_uw = np.array([0, 0])

        """define outliers"""
        #         new_arr = self.__smooth_list_gaussian(self.raw_frame[self.raw_frame['x_value'].isin(xx)
        #                                                             ].heat_value.values, degree=3)
        q75, q25 = np.percentile(self.raw_frame[self.raw_frame['x_value'].isin(xx)
                                 ].heat_value.values, [q_max, q_min])
        iqr = q75 - q25
        lower_whisker = q25 - 1.5 * iqr
        upper_whisker = q75 + 1.5 * iqr

        lower_mask = self.raw_frame['heat_value'].values < lower_whisker
        upper_mask = self.raw_frame['heat_value'].values > upper_whisker

        lw_outlier_cnt = len(self.raw_frame['heat_value'].values[lower_mask])
        up_outlier_cnt = len(self.raw_frame['heat_value'].values[upper_mask])

        lw_len = len(self.raw_frame['heat_value'].values[lower_mask])
        uw_len = len(self.raw_frame['heat_value'].values[upper_mask])

        """split outliers in data"""
        # lower whisker
        if lw_len > 1:
            xmin_lw = min(self.raw_frame['heat_value'].values[lower_mask])
            xmax_lw = max(self.raw_frame['heat_value'].values[lower_mask])
            x_lw = np.linspace(xmin_lw, xmax_lw, 1000)  # get 1000 points on x axis

            # get actual kernel density.
            density_lw = stats.gaussian_kde(self.raw_frame['heat_value'].values[lower_mask].flatten())
            y_lw = np.array(density_lw(x_lw))

        # upper whisker
        if uw_len > 1:
            xmin_uw = min(self.raw_frame['heat_value'].values[upper_mask])
            xmax_uw = max(self.raw_frame['heat_value'].values[upper_mask])
            x_uw = np.linspace(xmin_uw, xmax_uw, 1000)  # get 1000 points on x axis

            # get actual kernel density.
            density_uw = stats.gaussian_kde(self.raw_frame['heat_value'].values[upper_mask].flatten())
            y_uw = np.array(density_uw(x_uw))

        if (lw_len <= 1) and (uw_len <= 1):
            heat_min = self.glob_h_mi
            heat_max = self.glob_h_ma
            heat_sample = self.glob_samp
        else:
            """comparing outliers"""
            x = x_lw if max(np.fabs(y_lw)) > max(np.fabs(y_uw)) else x_uw
            y = y_lw if max(np.fabs(y_lw)) > max(np.fabs(y_uw)) else y_uw

            heat_min = xmin_lw if max(np.fabs(y_lw)) > max(np.fabs(y_uw)) else xmin_uw
            heat_max = xmax_lw if max(np.fabs(y_lw)) > max(np.fabs(y_uw)) else xmax_uw

            """getting heat sample"""
            heat_sample = x[list(y).index(min(y))]

        """info part"""
        if info:
            print('lower_whisker', lower_whisker)
            print('upper_whisker', upper_whisker)
            print('count of lower outliers', len(self.raw_frame['heat_value'].values[lower_mask]))
            print('count of upper outliers', len(self.raw_frame['heat_value'].values[upper_mask]))
            print('heat_sample', heat_sample)

        """plot part"""
        if plot:
            if list(x_lw) != [0, 0]:
                plt.plot(x_lw, y_lw, label='x = {}'.format(xx))
            if list(x_uw) != [0, 0]:
                plt.plot(x_uw, y_uw, label='x = {}'.format(xx))
            plt.legend()
            plt.show()

        return heat_sample, heat_min, heat_max

    def __smooth_list_gaussian(self, list1, degree=5):
        window = degree * 2 - 1
        weight = np.array([1.0] * window)
        weightGauss = []
        for i in range(window):
            i = i - degree + 1
            frac = i / float(window)
            gauss = 1 / (np.exp((4 * (frac)) ** 2))
            weightGauss.append(gauss)
        weight = np.array(weightGauss) * weight
        smoothed = [0.0] * (len(list1) - window)
        for i in range(len(smoothed)):
            smoothed[i] = sum(np.array(list1[i:i + window]) * weight) / sum(weight)
        return np.array(smoothed)

    def __find_nearest_universal(self, arr, value):
        array = np.asarray(arr)
        idx = (np.abs(array - value)).argmin()
        return array[idx]

    def __find_min_max_y_on_(self, x, heat_sample, wrong_y=None):

        n_steps = self.__n_steps

        if len(self.raw_frame[self.raw_frame.x_value == x].y_value) <= 0.7 * len(self.y_list):
            smooth_heat = self.__smooth_list_gaussian(self.raw_frame[self.raw_frame.x_value == x
                                                                     ].heat_value.values, degree=1)
        else:
            smooth_heat = self.__smooth_list_gaussian(self.raw_frame[self.raw_frame.x_value == x
                                                                     ].heat_value.values, degree=7)

        smooth_y = np.linspace(self.y_min, self.y_max, len(smooth_heat))

        if wrong_y is not None:
            mask = (smooth_y > wrong_y - n_steps * self.y_step) & (smooth_y < wrong_y + n_steps * self.y_step)
            smooth_heat[mask] = self.check_heat_samp(q_max=70, q_min=35)

        idx = list(smooth_heat).index(self.__find_nearest_universal(smooth_heat, heat_sample))
        y_min = smooth_y[idx] - n_steps * self.y_step
        y_max = smooth_y[idx] + n_steps * self.y_step

        return y_min, y_max

    def __change_outlier_heatsamples(self):
        stdev = np.std(self.heat_samples)
        # print(stats.mode(self.heat_samples))
        mode = np.percentile(self.heat_samples, 50)

        outliers_mask = np.abs(self.heat_samples - mode) > stdev

        self.heat_samples[outliers_mask] = np.mean(self.heat_samples[~outliers_mask])
        self.heat_mins[outliers_mask] = np.mean(self.heat_mins[~outliers_mask])
        self.heat_maxs[outliers_mask] = np.mean(self.heat_maxs[~outliers_mask])

    def __find_nearest(self, value):
        array = np.asarray(self.y_list)
        idx = (np.abs(array - value)).argmin()
        return array[idx]

    @property
    def interval(self):
        raw_len = len(self.raw_array[1, :])

        current_count = self.__find_nearest_universal(self.count_of_elements, raw_len)
        idx = list(self.count_of_elements).index(current_count)
        return self.timing[idx]



from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

setup_args = dict(
    name='pyquac',
    version='1.1.11',
    description='Useful tools for quantum computing experiments, provided for BMSTU FMN',
    long_description_content_type="text/markdown",
    long_description=README + '\n\n' + HISTORY,
    packages=find_packages(),
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: OS Independent'
    ],
    author='Nikolay Zhitkov',
    author_email='nokolay.zh@gmail.com',
    keywords=['Two tone spectroscopy', 'plotly', 'pandas'],
    url='https://github.com/ikaryss/pyquac',
    download_url='https://pypi.org/project/pyquac/'
)

# install_requires = [
#     'numpy>=1.20.0',
#     'matplotlib',
#     'scipy',
#     'pandas>=1.3.0',
#     'kaleido==0.1.0',
#     'dash>=1.21.0',
#     'jupyter_dash>=0.4.0',
#     'numba>=0.48.0',
#     'Pillow',
#     'typing',
#     'dash_core_components',
#     'dash_html_components',
#     'plotly',
#     'notebook',
#     'ipywidgets>=7.6.0',
# ]

install_requires = [
    'dash>=1.21.0',
    'jupyter_dash>=0.4.0',
    'numba>=0.54.0',
    'kaleido==0.1.0',
    'Pillow'
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)
